import os
from pathlib import Path
from typing import Literal, TypeVar, Self
import json
from logging import getLogger

from .task_queue import TaskQueue

logger = getLogger(__name__)
Bot = TypeVar("Bot")
BackTestEngine = TypeVar("BackTestEngine")
BackTestController = TypeVar("BackTestController")


class Config:
    login: int
    trade_record_mode: Literal["csv", "json"]
    password: str
    config_file: str | Path
    server: str
    path: str | Path
    timeout: int
    filename: str
    state: dict
    root: Path
    record_trades: bool
    records_dir: Path
    backtest_dir: Path
    task_queue: TaskQueue
    _backtest_engine: BackTestEngine
    bot: Bot
    backtest_controller: BackTestController
    _instance: Self
    mode: Literal["backtest", "live"]
    use_terminal_for_backtesting: bool
    shutdown: bool
    force_shutdown: bool
    _defaults = {
        "timeout": 60000,
        "record_trades": True,
        "config_file": None,
        "trade_record_mode": "csv",
        "mode": "live",
        "filename": "aiomql.json",
        "use_terminal_for_backtesting": True,
        "path": "",
        "login": 0,
        "password": "",
        "server": "",
        "shutdown": False,
        "force_shutdown": False,
        "root": None,
    }

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "_instance"):
            cls._instance = super().__new__(cls)
            cls._instance.state = {}
            cls._instance.task_queue = TaskQueue(mode='infinite', workers=10)
            cls._instance.set_attributes(**cls._defaults)
            cls._instance._backtest_engine = None
            cls._instance.bot = None
            cls._instance.backtest_controller = None
            # cls._instance.load_config(**kwargs)
        return cls._instance

    def __init__(self, **kwargs):
        """Initialize the Config object. The root directory can be set here or in the load_config method."""
        root = kwargs.pop("root", None)
        config_file = kwargs.pop("config_file", None)
        if self.root is None or root is not None or config_file is not None:
            self.load_config(root=root, config_file=config_file, **kwargs)
        else:
            self.set_attributes(**kwargs)

    @property
    def backtest_engine(self):
        """Returns the backtest engine object"""
        return self._backtest_engine

    @backtest_engine.setter
    def backtest_engine(self, value: BackTestEngine):
        """Set the backtest engine object"""
        self._backtest_engine = value

    def set_attributes(self, **kwargs):
        """Set keyword arguments as object attributes. The root folder attribute can't be set here.

        Args:
            **kwargs: Object attributes and values as keyword arguments
        """
        if kwargs.get("root", None) is not None:
            kwargs.pop("root", None)
            logger.debug("Tried setting root from set_attributes. Use load_config to change project root")
        if kwargs.get("config_file", None) is not None:
            kwargs.pop("config_file", None)
            logger.debug("Tried setting config_file from set_attributes. Use load_config to change project root")
        [setattr(self, key, value) for key, value in kwargs.items()]

    def find_config_file(self):
        try:
            current = Path.cwd()
            current = os.path.commonpath([current, self.root])
            current = Path(current).resolve()
            config_file = current / self.filename
            if config_file.exists():
                return config_file
            if current == self.root:
                return
            for dirname in current.parents:
                config_file = dirname / self.filename
                if config_file.exists():
                    return config_file

                if self.root == dirname:
                    break
        except Exception as err:
            logger.debug(f"Error finding config file: {err}")


    def load_config(self, *, config_file: str | Path = None, filename: str = None, root: str | Path = None, **kwargs) -> Self:
        """Load configuration settings from a file and reset the config object.

        Args:
            config_file (str | Path): The absolute path to the config file.
            filename (str): The name of the file to load if file path is not specified. If not provided aiomql.json is used
            root (str): The root directory of the project.
            **kwargs: Additional keyword arguments to be set on the config object.
        """
        if root is not None:
            root = Path(root).resolve()
            root.mkdir(parents=True, exist_ok=True) if not root.exists() else ...
            self.root = root
        else:
            self.root = self.root if isinstance(self.root, Path) else Path.cwd()

        if config_file is not None:
            config_file = Path(config_file).resolve()
            if not config_file.exists():
                self.filename = filename or self.filename
                self.config_file = self.find_config_file()
            else:
                self.filename = config_file.name
                self.config_file = config_file
        else:
            self.filename = filename or self.filename
            self.config_file = self.find_config_file()

        if self.config_file is None:
            logger.debug("No Config File Found")
            file_config = {}
        else:
            fh = open(self.config_file, mode="r")
            file_config = json.load(fh)
            fh.close()

        data = file_config | kwargs
        self.set_attributes(**data)

        try:
            if self.path:
                self.path = self.root / self.path if not Path(self.path).resolve().exists() else self.path
        except Exception as err:
            logger.debug(f"Error setting path: {err}")
            self.path = ""

        return self

    @property
    def records_dir(self):
        rec_dir = self.root / 'trade_records'
        rec_dir.mkdir(parents=True, exist_ok=True) if rec_dir.exists() is False else ...
        return rec_dir

    @property
    def backtest_dir(self) -> Path:
        b_dir = self.root / 'backtesting'
        b_dir.mkdir(parents=True, exist_ok=True) if b_dir.exists() is False else ...
        return b_dir

    def account_info(self) -> dict[str, int | str]:
        """Returns Account login details as found in the config object if available

        Returns:
            dict[str, int | str]: A dictionary of login details
        """
        return {"login": self.login, "password": self.password, "server": self.server}


Config.__doc__ = """A class for handling configuration settings for the aiomql package.
    Attributes:
        login (int): The account login number
        trade_record_mode (Literal["csv", "json"]): The mode for recording trades
        password (str): The account password
        server (str): The account server
        path (str | Path): The path to the terminal
        timeout (int): The timeout argument for the terminal
        filename (str): The filename of the config file
        config_file (Path): The config file path
        state (dict): The
        root (Path): The root directory of the project
        record_trades (bool): To record trades or not. Default is True
        records_dir (Path): The directory to store trade records, relative to the root directory
        backtest_dir (Path): The directory to store backtest results, relative to the root directory
        task_queue (TaskQueue): The TaskQueue object for handling background tasks
        _backtest_engine (BackTestEngine): The backtest engine object
        bot (Bot): The bot object
        _instance (Self): The instance of the Config class
        mode (Literal["backtest", "live"]): The trading mode, either backtest or live, default is live
        use_terminal_for_backtesting (bool): Use the terminal for backtesting, default is True
        shutdown (bool): A signal to shut down the terminal, default is False
        force_shutdown (bool): A signal to force shut down the terminal, default is False

    Notes:
        By default, the config class looks for a file named aiomql.json. This can be changed by setting the filename
        attribute to the desired file name. The root directory of the project can be set by passing the root argument
        to the load_config method or during object instantiation. If not provided it is assumed to be the current working
        directory. All directories and files are assumed to be relative to the root directory except when an absolute path
        is provided, this includes the config file, the records_dir and the backtest_dir attributes.
        The root directory is used to locate the config file and to set the records_dir and backtest_dir attributes.
    """



from .get_data import GetData, BackTestData
from .backtest_engine import BackTestEngine
from .backtest_account import BackTestAccount
from .trades_manager import PositionsManager, OrdersManager, DealsManager
from .backtest_controller import BackTestController



import asyncio
import time
from typing import Coroutine, Callable, Literal
from logging import getLogger
from signal import SIGINT, signal

logger = getLogger(__name__)


class QueueItem:
    """A class to represent a task item in the queue.

    Attributes:
        - `task_item` (Callable | Coroutine): The task to run.

        - `args` (tuple): The arguments to pass to the task

        - `kwargs` (dict): The keyword arguments to pass to the task

        - `must_complete` (bool): A flag to indicate if the task must complete before the queue stops. Default is False.

        - `time` (int): The time the task was added to the queue.
    """

    def __init__(self, task_item: Callable | Coroutine, *args, **kwargs):
        self.task_item = task_item
        self.args = args
        self.kwargs = kwargs
        self.must_complete = False
        self.time = time.monotonic_ns()

    def __hash__(self):
        return id(self)

    def __lt__(self, other):
        return self.time < other.time

    async def run(self):
        try:
            if asyncio.iscoroutinefunction(self.task_item):
                await self.task_item(*self.args, **self.kwargs)
            else:
                await asyncio.to_thread(self.task_item, *self.args, **self.kwargs)
        except Exception as err:
            logger.error("Error %s occurred in %s with args %s and %s",
                         err, self.task_item.__name__, self.args, self.kwargs)


class TaskQueue:
    queue_task: asyncio.Task

    def __init__(self, size: int = 0, workers: int = 500, timeout: int = None, queue: asyncio.Queue = None,
                 on_exit: Literal['cancel', 'complete_priority'] = 'complete_priority',
                 mode: Literal['finite', 'infinite'] = 'finite', worker_timeout: int = 60):

        self.queue = queue or asyncio.PriorityQueue(maxsize=size)
        self.workers = workers
        self.worker_tasks = []
        self.priority_tasks = set()  # tasks that must complete
        self.timeout = timeout
        self.stop = False
        self.on_exit = on_exit
        self.mode = mode
        self.worker_timeout = worker_timeout
        signal(SIGINT, self.sigint_handle)

    def add(self, *, item: QueueItem, priority=3, must_complete=False):
        """Add a task to the queue.

        Args:
            item (QueueItem): The task to add to the queue.
            priority (int): The priority of the task. Default is 3.
            must_complete (bool): A flag to indicate if the task must complete before the queue stops. Default is False.
        """
        try:
            if self.stop:
                return
            item.must_complete = must_complete
            self.priority_tasks.add(item) if item.must_complete else ...
            if isinstance(self.queue, asyncio.PriorityQueue):
                item = (priority, item)
            self.queue.put_nowait(item)
        except asyncio.QueueFull:
            logger.error("Queue is full")

    async def worker(self):
        """Worker function to run tasks in the queue."""
        while True:
            try:
                if self.mode == 'infinite' and self.queue.qsize() <= 1:
                    dummy = QueueItem(self.dummy_task)
                    self.add(item=dummy)

                if isinstance(self.queue, asyncio.PriorityQueue):
                    _, item = self.queue.get_nowait()
                else:
                    item = self.queue.get_nowait()

                if self.stop is False or item.must_complete:
                    await item.run()

                self.queue.task_done()

                self.priority_tasks.discard(item)

                if self.stop and (self.on_exit == 'cancel' or len(self.priority_tasks) == 0):
                    self.cancel()
                    break

            except asyncio.QueueEmpty:
                if self.stop:
                    break

                if self.mode == 'finite':
                    self.stop = True
                    break
            except Exception as err:
                logger.error("%s: Error occurred in worker", err)
                break

    async def dummy_task(self):
        await asyncio.sleep(self.worker_timeout)

    async def run(self, timeout: int = 0):
        """Run the queue until all tasks are completed or the timeout is reached.

        Args:
            timeout (int): The maximum time to wait for the queue to complete. Default is 0. If timeout is provided
                the queue is joined using `asyncio.wait_for` with the timeout. If the timeout is reached, the queue is
                stopped and the remaining tasks are handled based on the `on_exit` attribute.
                If the timeout is 0, the queue will run until all tasks are completed or the queue is stopped.
        """
        start = time.perf_counter()
        try:
            self.worker_tasks.extend(asyncio.create_task(self.worker()) for _ in range(self.workers))
            timeout = timeout or self.timeout
            self.queue_task = asyncio.create_task(self.queue.join())

            if timeout:
                await asyncio.wait_for(self.queue_task, timeout=timeout)

            else:
                await self.queue_task

        except TimeoutError:
            logger.warning("Timed out after %d seconds, %d tasks remaining",
                           time.perf_counter() - start, self.queue.qsize())
            self.stop = True
            await self.clean_up()

        except asyncio.CancelledError:
            self.stop = True
            await self.clean_up()

        except Exception as err:
            logger.warning("%s: An error occurred in %s.run", err, self.__class__.__name__)
            self.stop = True
            await self.clean_up()


    async def clean_up(self):
        """Clean up tasks in the queue, completing priority tasks if `on_exit` is `complete_priority`"""
        self.stop = True
        try:
            logger.info('cleaning up tasks...')
            if self.on_exit == 'complete_priority' and (pt := len(self.priority_tasks)) > 0:
                logger.info('Completing %d priority tasks...', pt)
                self.queue_task = asyncio.create_task(self.queue.join())
                await self.queue_task
            logger.info('Cleaning up tasks done...')
            self.cancel()

        except asyncio.CancelledError:
            self.cancel()

        except Exception as err:
            logger.error(f"%s: Error occurred in %s", err, self.__class__.__name__)
            self.cancel()

    def cancel(self):
        """Cancel all tasks in the queue"""
        try:
           self.queue_task.cancel()
        except asyncio.CancelledError:
            ...

        except Exception as err:
            logger.error("%s: occurred in canceling all tasks", err)

    def sigint_handle(self, sig, frame):
        logger.info('SIGINT received, cleaning up...')
        if self.stop is False:
            self.stop = True
        else:
            self.stop = True
            self.cancel()


TaskQueue.__doc__ = """TaskQueue is a class that allows you to queue tasks and run them concurrently with a specified number of workers.

    Attributes:
        - `workers` (int): The number of workers to run concurrently. Default is 10.

        - `timeout` (int): The maximum time to wait for the queue to complete. Default is None. If timeout is provided
            the queue is joined using `asyncio.wait_for` with the timeout.

        - `queue` (asyncio.Queue): The queue to store the tasks. Default is `asyncio.PriorityQueue` with no size limit.

        - `on_exit` (Literal["cancel", "complete_priority"]): The action to take when the queue is stopped.

        - `mode` (Literal["finite", "infinite"]): The mode of the queue. If `finite` the queue will stop when all tasks
            are completed. If `infinite` the queue will continue to run until stopped.

        - `worker_timeout` (int): The time to wait for a task to be added to the queue before stopping the worker or
            adding a dummy sleep task to the queue.

        - `stop` (bool): A flag to stop the queue instance.

        - `tasks` (list): A list of the worker tasks running concurrently, including the main task that joins the queue.

        - `priority_tasks` (set): A set to store the QueueItems that must complete before the queue stops.
    """



from datetime import datetime
from logging import getLogger
from typing import Literal, TypeVar

from numpy import ndarray
from MetaTrader5 import (
    Tick,
    SymbolInfo,
    AccountInfo,
    TerminalInfo,
    TradeOrder,
    TradePosition,
    TradeDeal,
    OrderCheckResult,
    OrderSendResult,
)

from .meta_trader import MetaTrader
from .constants import TimeFrame, CopyTicks, OrderType
from .._utils import error_handler

logger = getLogger(__name__)

BackTestEngine = TypeVar("BackTestEngine")


class MetaBackTester(MetaTrader):
    """A class for testing trading strategies in the MetaTrader 5 terminal. A subclass of MetaTrader.

    Attributes:
        backtest_engine (BackTestEngine): The backtesting engine to use for testing trading strategies.
    """

    backtest_engine: BackTestEngine

    def __init__(self, *, backtest_engine: BackTestEngine = None):
        super().__init__()
        self.backtest_engine = backtest_engine

    @property
    def backtest_engine(self) -> BackTestEngine:
        return self.config.backtest_engine

    @backtest_engine.setter
    def backtest_engine(self, value: BackTestEngine):
        if value is not None:
            self.config.backtest_engine = value

    async def last_error(self) -> tuple[int, str]:
        if self.config.use_terminal_for_backtesting is False:
            return -1, ""
        else:
            return await super().last_error()

    async def initialize(
        self,
        *,
        path: str = "",
        login: int = 0,
        password: str = "",
        server: str = "",
        timeout: int | None = None,
        portable=False,
    ) -> bool:
        if self.config.use_terminal_for_backtesting:
            return await super().initialize(path=path, login=login, password=password, server=server, timeout=timeout)

        return True

    def initialize_sync(
        self,
        *,
        path: str = "",
        login: int = 0,
        password: str = "",
        server: str = "",
        timeout: int | None = None,
        portable=False,
    ) -> bool:
        if self.config.use_terminal_for_backtesting:
            return super().initialize_sync(path=path, login=login, password=password, server=server, timeout=timeout)

        return True

    def login_sync(self, *, login: int = 0, password: str = "", server: str = "", timeout: int = 60000) -> bool:
        if self.config.use_terminal_for_backtesting:
            return super().login_sync(login=login, password=password, server=server, timeout=timeout)
        return True

    async def login(self, *, login: int = 0, password: str = "", server: str = "", timeout: int = 60000) -> bool:
        if self.config.use_terminal_for_backtesting:
            return await super().login(login=login, password=password, server=server, timeout=timeout)
        return True

    async def shutdown(self) -> None:
        await super().shutdown() if self.config.use_terminal_for_backtesting else ...

    @error_handler(msg="test data not available", exe=AttributeError)
    async def terminal_info(self) -> TerminalInfo:
        res = await self.backtest_engine.get_terminal_info()
        return res

    @error_handler(msg="test data not available", exe=AttributeError)
    async def account_info(self) -> AccountInfo:
        """"""
        return self.backtest_engine.get_account_info()

    @error_handler(msg="test data not available", exe=AttributeError)
    async def symbols_total(self) -> int:
        tot = await self.backtest_engine.get_symbols_total()
        return tot

    @error_handler(msg="test data not available", exe=AttributeError)
    async def symbols_get(self, group: str = "") -> tuple[SymbolInfo, ...] | None:
        """"""
        syms = await self.backtest_engine.get_symbols(group=group)
        return syms

    @error_handler(msg="test data not available")
    async def symbol_info(self, symbol: str) -> SymbolInfo | None:
        sym = await self.backtest_engine.get_symbol_info(symbol=symbol)
        return sym

    @error_handler(msg="test data not available")
    async def symbol_info_tick(self, symbol: str) -> Tick | None:
        tick = await self.backtest_engine.get_symbol_info_tick(symbol=symbol)
        return tick

    @error_handler(msg="test data not available", exe=AttributeError)
    async def copy_rates_from(
        self, symbol: str, timeframe: TimeFrame, date_from: datetime | float, count: int
    ) -> ndarray | None:
        rates = await self.backtest_engine.get_rates_from(
            symbol=symbol, timeframe=timeframe, date_from=date_from, count=count
        )
        return rates

    @error_handler(msg="test data not available", exe=AttributeError)
    async def copy_rates_from_pos(
        self, symbol: str, timeframe: TimeFrame, start_pos: int, count: int
    ) -> ndarray | None:
        rates = await self.backtest_engine.get_rates_from_pos(
            symbol=symbol, timeframe=timeframe, start_pos=start_pos, count=count
        )
        return rates

    @error_handler(msg="test data not available", exe=AttributeError)
    async def copy_rates_range(
        self, symbol: str, timeframe: TimeFrame, date_from: datetime | float, date_to: datetime | float
    ) -> ndarray | None:
        rates = await self.backtest_engine.get_rates_range(
            symbol=symbol, timeframe=timeframe, date_from=date_from, date_to=date_to
        )
        return rates

    @error_handler(msg="test data not available", exe=AttributeError)
    async def copy_ticks_from(
        self, symbol: str, date_from: datetime | float, count: int, flags: CopyTicks
    ) -> ndarray | None:
        ticks = await self.backtest_engine.get_ticks_from(symbol=symbol, date_from=date_from, count=count, flags=flags)
        return ticks

    @error_handler(msg="test data not available", exe=AttributeError)
    async def copy_ticks_range(
        self, symbol: str, date_from: datetime | float, date_to: datetime | float, flags: CopyTicks
    ) -> ndarray | None:
        ticks = await self.backtest_engine.get_ticks_range(
            symbol=symbol, date_from=date_from, date_to=date_to, flags=flags
        )
        return ticks

    @error_handler(msg="test data not available", exe=AttributeError)
    async def orders_total(self) -> int:
        return self.backtest_engine.get_orders_total()

    @error_handler(msg="test data not available", exe=AttributeError)
    async def orders_get(self, group: str = "", ticket: int = 0, symbol: str = "") -> tuple[TradeOrder, ...] | None:
        kwargs = {key: value for key, value in (("group", group), ("ticket", ticket), ("symbol", symbol)) if value}
        return self.backtest_engine.get_orders(**kwargs)

    @error_handler(msg="test data not available", exe=AttributeError)
    async def order_calc_margin(self, action: OrderType, symbol: str, volume: float, price: float) -> float | None:
        res = await self.backtest_engine.order_calc_margin(action=action, symbol=symbol, volume=volume, price=price)
        return res

    @error_handler(msg="test data not available", exe=AttributeError)
    async def order_calc_profit(
        self, action: Literal[0, 1], symbol: str, volume: float, price_open: float, price_close: float
    ) -> float | None:
        profit = await self.backtest_engine.order_calc_profit(
            action=action, symbol=symbol, volume=volume, price_open=price_open, price_close=price_close
        )
        return profit

    @error_handler(msg="test data not available", exe=AttributeError)
    async def order_check(self, request: dict) -> OrderCheckResult:
        ocr = await self.backtest_engine.order_check(request=request)
        return ocr

    async def order_send(self, request: dict) -> OrderSendResult:
        osr = await self.backtest_engine.order_send(request=request)
        return osr

    @error_handler(msg="test data not available", exe=AttributeError)
    async def positions_total(self) -> int:
        return self.backtest_engine.get_positions_total()

    @error_handler(msg="test data not available", exe=AttributeError)
    async def positions_get(
        self, group: str = "", ticket: int = None, symbol: str = ""
    ) -> tuple[TradePosition, ...] | None:
        kwargs = {key: value for key, value in (("group", group), ("ticket", ticket), ("symbol", symbol)) if value}
        return self.backtest_engine.get_positions(**kwargs)

    @error_handler(msg="test data not available", exe=AttributeError)
    async def history_orders_total(self, date_from: datetime | float, date_to: datetime | float) -> int | None:
        return self.backtest_engine.get_history_orders_total(date_from=date_from, date_to=date_to)

    @error_handler(msg="test data not available", exe=AttributeError)
    async def history_orders_get(
        self,
        date_from: datetime | float = None,
        date_to: datetime | float = None,
        group: str = "",
        ticket: int = None,
        position: int = None,
    ) -> tuple[TradeOrder, ...] | None:
        args = (
            ("date_from", date_from),
            ("date_to", date_to),
            ("group", group),
            ("ticket", ticket),
            ("position", position),
        )
        kwargs = {key: value for key, value in args if value}
        return self.backtest_engine.get_history_orders(**kwargs)

    @error_handler(msg="test data not available", exe=AttributeError)
    async def history_deals_total(self, date_from: datetime | float, date_to: datetime | float) -> int | None:
        return self.backtest_engine.get_history_deals_total(date_from=date_from, date_to=date_to)

    @error_handler(msg="test data not available", exe=AttributeError)
    async def history_deals_get(
        self,
        date_from: datetime | float = None,
        date_to: datetime | float = None,
        group: str = "",
        ticket: int = None,
        position: int = None,
    ) -> tuple[TradeDeal, ...] | None:
        args = (
            ("date_from", date_from),
            ("date_to", date_to),
            ("group", group),
            ("ticket", ticket),
            ("position", position),
        )
        kwargs = {key: value for key, value in args if value}
        return self.backtest_engine.get_history_deals(**kwargs)



import random
from logging import getLogger

from ...lib.strategy import Strategy
from ..symbols import ForexSymbol
from ..utils import Tracker
from ...core.constants import TimeFrame, OrderType
from ..traders.scalp_trader import ScalpTrader

logger = getLogger(__name__)


class Chaos(Strategy):
    """A chaotic strategy that buys and sells randomly."""

    ltf: TimeFrame
    htf: TimeFrame
    lcc: int
    hcc: int
    fast_ema: int
    slow_ema: int
    tracker: Tracker
    parameters = {"fast_ema": 8, "slow_ema": 20, "ltf": TimeFrame.M1, "htf": TimeFrame.M2, "lcc": 100, "hcc": 100}

    def __init__(self, *, symbol: ForexSymbol, params: dict = None, sessions=None, name="Chaos"):
        super().__init__(symbol=symbol, params=params, sessions=sessions, name=name)
        self.tracker = Tracker(snooze=self.ltf.seconds)
        self.trader = ScalpTrader(symbol=self.symbol)

    async def check_trend(self):
        try:
            candles = await self.symbol.copy_rates_from_pos(timeframe=self.htf, count=self.hcc)
            if (
                (current := candles[-1])
                and current.time < self.tracker.trend_time
                and current.close == self.tracker.last_trend_price
            ):
                self.tracker.update(new=False, order_type=None, snooze=5)
                return
            self.tracker.update(new=True, trend_time=current.time, last_trend_price=current.close)
            candles.ta.ema(length=self.slow_ema, append=True, fillna=0)
            candles.ta.ema(length=self.fast_ema, append=True, fillna=0)
            candles.rename(inplace=True, **{f"EMA_{self.fast_ema}": "fast", f"EMA_{self.slow_ema}": "slow"})
            order_type = random.choice([OrderType.BUY, OrderType.SELL])
            if order_type == OrderType.BUY:
                self.tracker.update(trend="bullish", snooze=self.htf.seconds, order_type=OrderType.BUY)
            else:
                self.tracker.update(trend="bearish", snooze=self.htf.seconds, order_type=OrderType.SELL)
        except Exception as err:
            logger.error(f"{err}. Failed to check trend")
            self.tracker.update(trend="ranging", snooze=self.ltf.seconds, order_type=None)

    async def trade(self):
        try:
            await self.check_trend()
            if self.tracker.order_type is not None:
                await self.trader.place_trade(order_type=self.tracker.order_type, parameters=self.parameters)
                self.tracker.update(order_type=None)
                await self.sleep(secs=self.tracker.snooze)
            else:
                await self.sleep(secs=self.tracker.snooze)
        except Exception as err:
            logger.error(f"{err}. Failed to trade {self.symbol.name} with {self.__class__.__name__}")



import asyncio
from datetime import datetime
from logging import getLogger
from typing import Literal, Self
from pathlib import Path

import numpy as np
from MetaTrader5 import (
    BookInfo,
    SymbolInfo,
    AccountInfo,
    Tick,
    TerminalInfo,
    TradeOrder,
    TradeDeal,
    TradePosition,
    OrderSendResult,
    OrderCheckResult,
)
import MetaTrader5 as mt5

from .constants import OrderType, CopyTicks

from ._core import MetaCore
from .errors import Error
from .config import Config

logger = getLogger()


class MetaTrader(MetaCore):
    def __init__(self):
        self.config = Config()
        self.error: Error = Error(1)

    async def __aenter__(self) -> Self:
        """
        Async context manager entry point.
        Initializes the connection to the MetaTrader terminal.

        Returns:
            MetaTrader: An instance of the MetaTrader class.
        """
        await self.initialize()
        await self.login()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Async context manager exit point. Closes the connection to the MetaTrader terminal.
        """
        await self.shutdown()

    async def _handler(self, api: dict):
        func = api["func"]
        args = api.get("args", ())
        kwargs = api.get("kwargs", {})
        error_msg = api.get("error_msg", "An error occurred")

        res = await asyncio.to_thread(func, *args, **kwargs)
        if res is None:
            err = await self.last_error()
            self.error = Error(*err)

            if self.error.is_connection_error():
                await self.initialize()
                await self.login()
                res = await asyncio.to_thread(func, *args, **kwargs)

                if res is None:
                    err = await self.last_error()
                    self.error = Error(*err)
                    logger.warning(f"{error_msg}:{self.error.description}")
            else:
                logger.warning(f"{error_msg}:{self.error.description}")
        return res

    async def login(self, *, login: int = 0, password: str = "", server: str = "", timeout: int = 60000) -> bool:
        """
        Connects to the MetaTrader terminal using the specified login, password and server.

        Args:
            login (int): The trading account number.
            password (str): The trading account password.
            server (str): The trading server name.
            timeout (int): The timeout for the connection in seconds.

        Returns:
            bool: True if successful, False otherwise.
        """
        acc_details = self.config.account_info()
        login = login or acc_details.get("login", 0)
        password = password or acc_details.get("password", "")
        server = server or acc_details.get("server", "")
        return await asyncio.to_thread(self._login, login, password=password, server=server, timeout=timeout)

    def login_sync(self, *, login: int = 0, password: str = "", server: str = "", timeout: int = 60000) -> bool:
        """
        Connects to the MetaTrader terminal using the specified login, password and server.

        Args:
            login (int): The trading account number.
            password (str): The trading account password.
            server (str): The trading server name.
            timeout (int): The timeout for the connection in seconds.

        Returns:
            bool: True if successful, False otherwise.
        """
        acc_details = self.config.account_info()
        login = login or acc_details.get("login", 0)
        password = password or acc_details.get("password", "")
        server = server or acc_details.get("server", "")
        res = mt5.login(login, password=password, server=server, timeout=timeout)
        return res

    async def initialize(
        self,
        path: str = None,
        login: int = 0,
        password: str = "",
        server: str = "",
        timeout: int | None = None,
        portable=False,
    ) -> bool:
        """
        Initializes the connection to the MetaTrader terminal. All parameters are optional.

        Keyword Args:
            path (str): The path to the MetaTrader terminal executable.
            login (int): The trading account number.
            password (str): The trading account password.
            server (str): The trading server name.
            timeout (int): The timeout for the connection in milliseconds.
            portable (bool): If True, the terminal will be launched in portable mode.

        Returns:
            bool: True if successful, False otherwise.
        """
        async with asyncio.Lock() as _:
            path = self.config.path if path is None else path
            path = "" if Path(path).exists() is False else path
            args = (str(path),) if path else ()
            acc = self.config.account_info()
            kwargs = {
                key: value
                for key, value in (
                    ("login", login or acc.get("login")),
                    ("password", password or acc.get("password")),
                    ("server", server or acc.get("server")),
                    ("timeout", timeout or 60000),
                    ("portable", portable),
                )
                if key is not None
            }
            res = await asyncio.to_thread(self._initialize, *args, **kwargs)
            if res is False:
                await self.shutdown()
                res = await asyncio.to_thread(self._initialize, *args, **kwargs)
            if not res:
                err = await self.last_error()
                self.error = Error(*err)
            return res

    def initialize_sync(
        self,
        path: str = None,
        login: int = 0,
        password: str = "",
        server: str = "",
        timeout: int | None = None,
        portable=False,
    ) -> bool:
        """
        Initializes the connection to the MetaTrader terminal. All parameters are optional.

        Keyword Args:
            path (str): The path to the MetaTrader terminal executable.
            login (int): The trading account number.
            password (str): The trading account password.
            server (str): The trading server name.
            timeout (int): The timeout for the connection in milliseconds.
            portable (bool): If True, the terminal will be launched in portable mode.

        Returns:
            bool: True if successful, False otherwise.
        """
        path = self.config.path if path is None else path
        path = "" if Path(path).exists() is False else path
        args = (str(path),) if path else ()
        acc = self.config.account_info()
        kwargs = {
            key: value
            for key, value in (
                ("login", login or acc.get("login")),
                ("password", password or acc.get("password")),
                ("server", server or acc.get("server")),
                ("timeout", timeout or 60000),
                ("portable", portable),
            )
            if key is not None
        }
        res = mt5.initialize(*args, **kwargs)
        if res is False:
            self._shutdown()
        if not res:
            err = self._last_error()
            self.error = Error(*err)
        return res

    async def shutdown(self) -> None:
        """Closes the connection to the MetaTrader terminal."""
        self._shutdown()

    async def last_error(self) -> tuple[int, str]:
        try:
            res = await asyncio.to_thread(self._last_error)
            return res
        except Exception as err:
            logger.warning(f"Error in obtaining last error.")
            return -1, str(err)

    async def version(self) -> tuple[int, int, str] | None:
        """"""
        api = {"func": self._version, "error_msg": "Error in obtaining version."}
        res = await self._handler(api)
        return res

    async def account_info(self) -> AccountInfo | None:
        """"""
        api = {"func": self._account_info, "error_msg": "Error in obtaining account information"}
        res = await self._handler(api)
        return res

    async def terminal_info(self) -> TerminalInfo | None:
        api = {"func": self._terminal_info, "error_msg": "Error in obtaining terminal information"}
        res = await self._handler(api)
        return res

    async def symbols_total(self) -> int:
        api = {"func": self._symbols_total, "error_msg": "Error in obtaining total symbols."}
        res = await self._handler(api)
        return res

    async def symbols_get(self, group: str = "") -> tuple[SymbolInfo] | None:
        kwargs = {"group": group} if group else {}
        api = {"func": self._symbols_get, "kwargs": kwargs, "error_msg": "Error in obtaining symbols."}
        res = await self._handler(api)
        return res

    async def symbol_info(self, symbol: str) -> SymbolInfo | None:
        api = {
            "func": self._symbol_info,
            "args": (symbol,),
            "error_msg": f"Error in obtaining information for {symbol}",
        }
        res = await self._handler(api)
        return res

    async def symbol_info_tick(self, symbol: str) -> Tick | None:
        api = {"func": self._symbol_info_tick, "args": (symbol,), "error_msg": f"Error in obtaining tick for {symbol}"}
        res = await self._handler(api)
        return res

    async def symbol_select(self, symbol: str, enable: bool) -> bool:
        api = {"func": self._symbol_select, "args": (symbol, enable), "error_msg": f"Error in selecting {symbol}"}
        res = await self._handler(api)
        return res

    async def market_book_add(self, symbol: str) -> bool:
        api = {
            "func": self._market_book_add,
            "args": (symbol,),
            "error_msg": f"Error in adding {symbol} to market book",
        }
        res = await self._handler(api)
        return res

    async def market_book_get(self, symbol: str) -> tuple[BookInfo] | None:
        api = {
            "func": self._market_book_get,
            "args": (symbol,),
            "error_msg": f"Error in obtaining market depth for {symbol}",
        }
        res = await self._handler(api)
        return res

    async def market_book_release(self, symbol: str) -> bool:
        api = {
            "func": self._market_book_release,
            "args": (symbol,),
            "error_msg": f"Error in releasing market depth for {symbol}",
        }
        res = await self._handler(api)
        return res

    async def copy_rates_from(
        self, symbol: str, timeframe: int, date_from: datetime | float, count: int
    ) -> np.ndarray | None:
        api = {
            "func": self._copy_rates_from,
            "args": (symbol, timeframe, date_from, count),
            "error_msg": f"Error in obtaining rates for {symbol}",
        }
        res = await self._handler(api)
        return res

    async def copy_rates_from_pos(self, symbol: str, timeframe: int, start_pos: int, count: int) -> np.ndarray | None:
        api = {
            "func": self._copy_rates_from_pos,
            "args": (symbol, timeframe, start_pos, count),
            "error_msg": f"Error in obtaining rates for {symbol}",
        }
        res = await self._handler(api)
        return res

    async def copy_rates_range(
        self, symbol: str, timeframe: int, date_from: datetime | float, date_to: datetime | float
    ) -> np.ndarray | None:
        api = {
            "func": self._copy_rates_range,
            "args": (symbol, timeframe, date_from, date_to),
            "error_msg": f"Error in obtaining rates for {symbol}",
        }
        res = await self._handler(api)
        return res

    async def copy_ticks_from(
        self, symbol: str, date_from: datetime | float, count: int, flags: CopyTicks
    ) -> np.ndarray | None:
        api = {
            "func": self._copy_ticks_from,
            "args": (symbol, date_from, count, flags),
            "error_msg": f"Error in obtaining ticks for {symbol}",
        }
        res = await self._handler(api)
        return res

    async def copy_ticks_range(
        self, symbol: str, date_from: datetime | float, date_to: datetime | float, flags: CopyTicks
    ) -> np.ndarray | None:
        api = {
            "func": self._copy_ticks_range,
            "args": (symbol, date_from, date_to, flags),
            "error_msg": f"Error in obtaining ticks for {symbol}",
        }
        res = await self._handler(api)
        return res

    async def orders_total(self) -> int:
        api = {"func": self._orders_total, "error_msg": "Error in obtaining total orders."}
        res = await self._handler(api)
        return res

    async def orders_get(self, group: str = "", ticket: int = 0, symbol: str = "") -> tuple[TradeOrder] | None:
        kwargs = {key: value for key, value in (("group", group), ("ticket", ticket), ("symbol", symbol)) if value}
        api = {"func": self._orders_get, "kwargs": kwargs, "error_msg": "Error in obtaining orders."}
        res = await self._handler(api)
        return res

    async def order_calc_margin(
        self, action: Literal[OrderType.BUY, OrderType.SELL], symbol: str, volume: float, price: float
    ) -> float | None:
        api = {
            "func": self._order_calc_margin,
            "args": (action, symbol, volume, price),
            "error_msg": "Error in calculating margin.",
        }
        res = await self._handler(api)
        return res

    async def order_calc_profit(
        self,
        action: Literal[OrderType.BUY, OrderType.SELL],
        symbol: str,
        volume: float,
        price_open: float,
        price_close: float,
    ) -> float | None:
        api = {
            "func": self._order_calc_profit,
            "args": (action, symbol, volume, price_open, price_close),
            "error_msg": "Error in calculating profit.",
        }
        res = await self._handler(api)
        return res

    async def order_check(self, request: dict) -> OrderCheckResult:
        api = {"func": self._order_check, "args": (request,), "error_msg": "Error in checking order."}
        res = await self._handler(api)
        return res

    async def order_send(self, request: dict) -> OrderSendResult:
        api = {"func": self._order_send, "args": (request,), "error_msg": "Error in sending order."}
        res = await self._handler(api)
        return res

    async def positions_total(self) -> int:
        api = {"func": self._positions_total, "error_msg": "Error in obtaining total positions."}
        res = await self._handler(api)
        return res

    async def positions_get(self, group: str = "", ticket: int = None, symbol: str = "") -> tuple[TradePosition] | None:
        kwargs = {key: value for key, value in (("group", group), ("ticket", ticket), ("symbol", symbol)) if value}
        api = {"func": self._positions_get, "kwargs": kwargs, "error_msg": "Error in obtaining open positions."}
        res = await self._handler(api)
        return res

    async def history_orders_total(self, date_from: datetime | float, date_to: datetime | float) -> int:
        api = {
            "func": self._history_orders_total,
            "args": (date_from, date_to),
            "error_msg": "Error in obtaining total history orders.",
        }
        res = await self._handler(api)
        return res

    async def history_orders_get(
        self,
        date_from: datetime | float = None,
        date_to: datetime | float = None,
        group: str = "",
        ticket: int = None,
        position: int = None,
    ) -> tuple[TradeOrder] | None:
        kwargs = {key: value for key, value in (("group", group), ("ticket", ticket), ("position", position)) if value}
        args = tuple(arg for arg in (date_from, date_to) if arg)
        api = {
            "func": self._history_orders_get,
            "args": args,
            "kwargs": kwargs,
            "error_msg": "Error in obtaining history orders",
        }
        res = await self._handler(api)
        return res

    async def history_deals_total(self, date_from: datetime | float, date_to: datetime | float) -> int:
        api = {
            "func": self._history_deals_total,
            "args": (date_from, date_to),
            "error_msg": "Error in obtaining total history deals",
        }
        res = await self._handler(api)
        return res

    async def history_deals_get(
        self,
        date_from: datetime | float = None,
        date_to: datetime | float = None,
        group: str = "",
        ticket: int = None,
        position: int = None,
    ) -> tuple[TradeDeal] | None:
        kwargs = {key: value for key, value in (("group", group), ("ticket", ticket), ("position", position)) if value}
        args = tuple(arg for arg in (date_from, date_to) if arg)
        api = {
            "func": self._history_deals_get,
            "args": args,
            "kwargs": kwargs,
            "error_msg": "Error in obtaining history deals",
        }
        res = await self._handler(api)
        return res



from logging import getLogger
from asyncio import Task
from signal import signal, SIGINT
from threading import Barrier, BrokenBarrierError
from typing import Self
from datetime import datetime

from ..config import Config
from ..exceptions import StopTrading

logger = getLogger(__name__)


class BackTestController:
    """The controller for the backtesting engine.
    It also acts as a synchronizer for running multiple strategies (tasks) using a threading.Barrier primitive.
    It handles the updating of open positions and close them when necessary.
    It handles the iterator for the backtesting engine and handles it movement in time by moving it to the next time step.

    Attributes:
        _instance (Self): The instance of the controller
        config (Config): The configuration for the backtesting engine
        tasks (list[Task]): The tasks that are being run
        barrier (Barrier): The barrier for synchronizing the tasks
    """

    _instance: Self
    config: Config
    tasks: list[Task]
    barrier: Barrier

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "_instance"):
            cls._instance = super().__new__(cls)
            cls._instance.config = Config()
            cls._instance.config.backtest_controller = cls._instance
            cls._instance.barrier = Barrier(1)
            cls._instance.tasks = []
        return cls._instance

    def __init__(self):
        signal(SIGINT, self.sigint_handler)

    @property
    def backtest_engine(self):
        """Returns the backtest engine"""
        return self.config.backtest_engine

    def add_tasks(self, *tasks: Task):
        """Adds tasks to the tasks list"""
        self.tasks.extend(tasks)

    def set_parties(self, *, parties: int):
        """Sets the number of parties for the barrier. The barrier will wait for the number of parties to reach the barrier.
        This has to be done here as it can be impossible to know the eventual number of parties to set the barrier to during initialization.

        Args:
            parties (int): The number of parties to set the barrier to
        """
        self.barrier._parties = parties

    @property
    def parties(self):
        """Returns the number of parties for the barrier"""
        return self.barrier.parties

    def sigint_handler(self, sig, frame):
        logger.warning("SIGINT received. Stopping the controller.")
        self.backtest_engine.stop_testing = True

    async def control(self):
        """The backtest controller. It controls the backtesting engine and the tasks that are being run.
        It acts as a synchronizer for the tasks and the backtesting engine.
        """
        try:
            self.backtest_engine.next()
            while True:
                pending = self.wait()
                # all main tasks have been completed in the current cycle
                if pending == 0:
                    await self.backtest_engine.tracker()
                    self.backtest_engine.next()
                    if self.backtest_engine.cursor.time % (3600 * 12) == 0:
                        logger.info(
                            datetime.strftime(
                                datetime.fromtimestamp(self.backtest_engine.cursor.time), "%Y-%m-%d %H:%M:%S"
                            )
                        )
                if self.backtest_engine.stop_testing:
                    logger.info(
                        "Stop trading called in control at %s",
                        datetime.fromtimestamp(self.backtest_engine.cursor.time).strftime("%Y-%m-%d %H:%M:%S"),
                    )
                    break
            await self.backtest_engine.wrap_up()
            self.stop_backtesting()
        except BrokenBarrierError:
            return

        except Exception as err:
            logger.error("Error: %s in controller", err)
            await self.backtest_engine.wrap_up()
            self.stop_backtesting()
            return

    def stop_backtesting(self):
        """Stop the backtester, and shutdown the executor"""
        self.abort()
        self.config.shutdown = True

    def wait(self):
        """Called by individual tasks to indicate completion of their cycle"""
        try:
            pending = self.barrier.wait()
            return pending
        except BrokenBarrierError:
            raise StopTrading
        except Exception as err:
            logger.error("Error: %s in wait", err)

    def abort(self):
        """Aborts the barrier"""
        self.barrier.abort()



from ...lib.symbol import Symbol


class ForexSymbol(Symbol):
    """Subclass of Symbol for Forex Symbols. Handles the conversion of currency and the computation of stop loss,
    take profit and volume.
    """

    @property
    def pip(self):
        """Returns the pip value of the symbol. This is ten times the point value for forex symbols.

        Returns:
            float: The pip value of the symbol.
        """
        return self.point * 10

    def compute_points(self, *, amount: float, volume: float) -> float:
        """Compute the number of points required for a trade. Given the amount and the volume of the trade.

        Args:
            amount (float): Amount to trade
            volume (float): Volume to trade
        """
        points = amount / (volume * self.point * self.trade_contract_size)
        return points

    async def compute_volume_points(self, *, amount: float, points: float, round_down: bool = False) -> float:
        """Compute the volume required for a trade. Given the amount and the number of points.

        Args:
            amount (float): Amount to trade
            points (float): Number of points
            round_down: round down the computed volume to the nearest step default True
        """
        volume = amount / (self.point * points * self.trade_contract_size)
        return self.round_off_volume(volume=volume, round_down=round_down)

    async def compute_volume_sl(self, *, amount: float, price: float, sl: float, round_down: bool = False) -> float:
        """Compute the volume required for a trade. Given the amount, the price and the stop loss.

        Args:
            amount (float): Amount to trade
            price (float): The price of the trade
            sl (float): The stop loss of the trade
            round_down (bool): round down the computed volume to the nearest step default to False

        Returns:
            float: The volume required for the trade
        """
        volume = amount / (abs(price - sl) * self.trade_contract_size)
        return self.round_off_volume(volume=volume, round_down=round_down)



import asyncio
import os
from concurrent.futures import ThreadPoolExecutor
from signal import signal, SIGINT
from typing import Coroutine, Callable
from logging import getLogger

from ..core.config import Config
from .strategy import Strategy

logger = getLogger(__name__)


class Executor:
    """Executor class for running multiple strategies on multiple symbols concurrently.

    Attributes:
        executor (ThreadPoolExecutor): The executor object.
        strategy_runners (list): List of strategies.
        coroutines (list[Coroutine]): A list of coroutines to run in the executor
        functions (dict[Callable, dict]): A dictionary of functions to run in the executor
    """

    executor: ThreadPoolExecutor
    tasks: list[asyncio.Task | asyncio.Future]
    config: Config

    def __init__(self):
        self.strategy_runners: list[Strategy] = []
        self.coroutines: list[Coroutine] = []
        self.coroutine_threads: list[Coroutine] = []
        self.functions: dict[Callable:dict] = {}
        self.tasks = []
        self.config = Config()
        self.timeout = None  # Timeout for the executor. For testing purposes only
        signal(SIGINT, self.sigint_handle)

    def add_function(self, *, function: Callable, kwargs: dict = None):
        kwargs = kwargs or {}
        self.functions[function] = kwargs

    def add_coroutine(self, *, coroutine: Callable | Coroutine, kwargs: dict = None, on_separate_thread=False):
        kwargs = kwargs or {}
        coroutine = coroutine(**kwargs)
        self.coroutines.append(coroutine) if on_separate_thread is False else self.coroutine_threads.append(coroutine)

    def add_strategies(self, *, strategies: tuple[Strategy]):
        """Add multiple strategies at once

        Args:
            strategies (Sequence[Strategy]): A sequence of strategies.
        """
        self.strategy_runners.extend(strategies)

    def add_strategy(self, *, strategy: Strategy):
        """Add a strategy instance to the list of workers

        Args:
            strategy (Strategy): A strategy object
        """
        self.strategy_runners.append(strategy)

    async def create_strategy_task(self, strategy: Strategy):
        task = asyncio.create_task(strategy.run_strategy())
        self.tasks.append(task)
        await task

    def run_strategy(self, strategy: Strategy):
        """Wraps the coroutine trade method of each strategy with 'asyncio.run'.

        Args:
            strategy (Strategy): A strategy object
        """
        asyncio.run(self.create_strategy_task(strategy))

    async def create_coroutine_task(self, coroutine: Coroutine):
        task = asyncio.create_task(coroutine)
        self.tasks.append(task)
        await task

    async def create_coroutines_task(self):
        """"""
        tasks = [asyncio.create_task(coroutine) for coroutine in self.coroutines]
        self.tasks.extend(tasks)
        task = asyncio.gather(*tasks, return_exceptions=True)
        self.tasks.append(task)
        await task

    def run_coroutine_tasks(self):
        """Run all coroutines in the executor"""
        asyncio.run(self.create_coroutines_task())

    def run_coroutine_task(self, coroutine):
        asyncio.run(self.create_coroutine_task(coroutine))

    @staticmethod
    def run_function(function: Callable, kwargs: dict):
        """Run a function

        Args:
            function: The function to run
            kwargs: A dictionary of keyword arguments for the function
        """
        try:
            function(**kwargs)
        except Exception as err:
            logger.error(f"Error: {err}. Unable to run function: {function.__name__}")

    def sigint_handle(self, signum, frame):
        self.config.shutdown = True

    async def exit(self):
        """Shutdown the executor"""
        start = asyncio.get_event_loop().time()
        try:
            while self.config.shutdown is False and self.config.force_shutdown is False:
                if self.timeout is not None and self.timeout < (asyncio.get_event_loop().time() - start):
                    self.config.shutdown = True
                    break
                timeout = self.timeout or 30
                await asyncio.sleep(timeout)
            for strategy in self.strategy_runners:
                strategy.running = False

            if self.config.backtest_engine is not None:
                self.config.backtest_engine.stop_testing = True

            self.executor.shutdown(wait=False, cancel_futures=False)

            for task in self.tasks:
                task.cancel()

            if self.config.force_shutdown:
                os._exit(1)
        except Exception as err:
            logger.error("%s: Unable to shutdown executor", err)
            os._exit(1)

    def execute(self, *, workers: int = 5):
        """Run the strategies with a threadpool executor.

        Args:
            workers: Number of workers to use in executor pool. Defaults to 5.

        Notes:
            No matter the number specified, the executor will always use a minimum of 5 workers.
        """
        workers_ = len(self.strategy_runners) + len(self.functions) + len(self.coroutine_threads) + 3
        workers = max(workers, workers_)
        with ThreadPoolExecutor(max_workers=workers) as executor:
            self.executor = executor
            [self.executor.submit(self.run_strategy, strategy) for strategy in self.strategy_runners]
            [self.executor.submit(function, **kwargs) for function, kwargs in self.functions.items()]
            [self.executor.submit(self.run_coroutine_task, coroutine) for coroutine in self.coroutine_threads]
            self.executor.submit(self.run_coroutine_tasks)



import asyncio
from datetime import datetime
from logging import getLogger

import pytz

from ..core.config import Config
from ..core.meta_trader import MetaTrader
from ..core.models import TradeDeal, TradeOrder
from ..core.meta_backtester import MetaBackTester
from .._utils import backoff_decorator

logger = getLogger(__name__)


class History:
    """The history class handles completed trade deals and trade orders in the trading history of an account.

    Attributes:
        deals (list[TradeDeal]): Iterable of trade deals
        orders (list[TradeOrder]): Iterable of trade orders
        total_deals: Total number of deals
        total_orders (int): Total number orders
        group (str): Filter for selecting history by symbols.
        mt5 (MetaTrader): MetaTrader instance
        config (Config): Config instance
    """

    mt5: MetaTrader | MetaBackTester
    config: Config
    deals: tuple[TradeDeal, ...]
    orders: tuple[TradeOrder, ...]
    total_deals: int
    total_orders: int
    group: str

    def __init__(
        self, *, date_from: datetime | float, date_to: datetime | float, group: str = "", use_utc: bool = True
    ):
        """
        Args:
            date_from (datetime, float): Date the orders are requested from. Set by the 'datetime' object or as a
                number of seconds elapsed since 1970.01.01. Defaults to twenty-four hours from the current time in 'utc'

            date_to (datetime, float): Date up to which the orders are requested. Set by the 'datetime' object or as a
                number of seconds elapsed since 1970.01.01. Defaults to the current time in "utc"

            group (str): Filter for selecting history by symbols. Defaults to an empty string
        """
        self.config = Config()
        self.mt5 = MetaTrader() if self.config.mode != "backtest" else MetaBackTester()
        date_from = date_from if isinstance(date_from, datetime) else datetime.fromtimestamp(date_from)
        date_to = date_to if isinstance(date_to, datetime) else datetime.fromtimestamp(date_to)
        self.date_from = date_from.astimezone(pytz.UTC) if use_utc else date_from
        self.date_to = date_to.astimezone(pytz.UTC) if use_utc else date_to
        self.group = group
        self.deals: tuple[TradeDeal, ...] = ()
        self.orders: tuple[TradeOrder, ...] = ()
        self.total_deals: int = 0
        self.total_orders: int = 0

    async def initialize(self):
        """Get history deals and orders"""
        deals, orders = await asyncio.gather(self.get_deals(), self.get_orders(), return_exceptions=True)
        self.deals = deals if isinstance(deals, tuple) else ()
        self.orders = orders if isinstance(orders, tuple) else ()
        self.total_deals = len(self.deals)
        self.total_orders = len(self.orders)

    @backoff_decorator
    async def get_deals(self) -> tuple[TradeDeal, ...]:
        """Get deals from trading history using the parameters set in the constructor.

        Returns:
            tuple[TradeDeal, ...]: A list of trade deals
        """
        deals = await self.mt5.history_deals_get(date_from=self.date_from, date_to=self.date_to, group=self.group)
        if deals is not None:
            return tuple(TradeDeal(**deal._asdict()) for deal in deals)
        logger.warning(f"Failed to get deals")
        return tuple()

    def get_deals_by_ticket(self, *, ticket: int) -> tuple[TradeDeal, ...]:
        """Call specifying the order ticket. Return all deals having the specified order ticket in the DEAL_ORDER
        property.

        Args:
            ticket (int): The order ticket

        Returns:
            tuple[TradeDeal]: A tuple of all deals with the order ticket
        """
        return tuple(sorted((deal for deal in self.deals if deal.order == ticket), key=lambda x: x.time_msc))

    def get_deals_by_position(self, *, position: int = None) -> tuple[TradeDeal, ...]:
        """
        Get all deals with the specified position ticket in the DEAL_POSITION_ID property
        Args:
            position (int): The position ticket

        Returns:
            tuple[TradeDeal]: A tuple of all deals with the position ticket
        """
        return tuple(sorted((deal for deal in self.deals if deal.position_id == position), key=lambda x: x.time_msc))

    @backoff_decorator
    async def get_orders(self) -> tuple[TradeOrder, ...]:
        """Get orders from trading history using the parameters set in the constructor or the method arguments.

        Returns:
            list[TradeOrder]: A list of trade orders
        """
        orders = await self.mt5.history_orders_get(date_from=self.date_from, date_to=self.date_to, group=self.group)

        if orders is not None:
            return tuple(TradeOrder(**order._asdict()) for order in orders)

        logger.warning(f"Failed to get orders")
        return tuple()

    def get_orders_by_ticket(self, *, ticket: int) -> tuple[TradeOrder, ...]:
        """filter orders by ticket"""
        return tuple(sorted((order for order in self.orders if order.ticket == ticket), key=lambda x: x.time_done_msc))

    def get_orders_by_position(self, *, position: int) -> tuple[TradeOrder, ...]:
        """filter orders by position"""
        return tuple(
            sorted((order for order in self.orders if order.position_id == position), key=lambda x: x.time_done_msc)
        )



from dataclasses import dataclass
from typing import ClassVar

from ..constants import AccountTradeMode, AccountMarginMode, AccountStopOutMode


@dataclass
class BackTestAccount:
    """Account data for backtesting"""

    login: int = 0
    trade_mode: AccountTradeMode = AccountTradeMode.DEMO
    leverage: float = 1
    limit_orders: float = 0
    margin_so_mode: AccountStopOutMode = AccountStopOutMode.PERCENT
    trade_allowed: bool = True
    trade_expert: bool = True
    margin_mode: AccountMarginMode = AccountMarginMode.EXCHANGE
    currency_digits: int = 2
    fifo_close: bool = False
    balance: float = 0
    credit: float = 0
    profit: float = 0
    equity: float = 0
    margin: float = 0
    margin_free: float = 0
    margin_level: float = 0
    margin_so_call: float = 0
    margin_so_so: float = 0
    margin_initial: float = 0
    margin_maintenance: float = 0
    assets: float = 0
    liabilities: float = 0
    commission_blocked: float = 0
    name: str = ""
    server: str = ""
    currency: str = "USD"
    company: str = ""

    __match_args__: ClassVar[tuple]

    def get_dict(self, exclude: set = None, include: set = None):
        """Returns a dictionary of the account data. Using the exclude and include arguments, you can filter the data

        Args:
            exclude (set): A set of keys to exclude
            include (set): A set of keys to include
        """
        exclude, include = exclude or set(), include or set()
        filter_ = include or set(self.__match_args__).difference(exclude)
        return {key: value for key, value in self.asdict().items() if key in filter_}

    def asdict(self):
        """Returns a dictionary of the account data"""
        res = {key: getattr(self, key) for key in self.__match_args__}
        return res

    def set_attrs(self, **kwargs):
        """Se the attributes of the account data to the instance"""
        [setattr(self, k, v) for k, v in kwargs.items() if k in self.__match_args__]



"""Handle Open positions."""

import asyncio
from logging import getLogger

from ..core.meta_trader import MetaTrader
from ..core.models import TradePosition, OrderSendResult
from ..core.constants import OrderType, TradeAction
from ..core.config import Config
from .._utils import backoff_decorator
from ..core.meta_backtester import MetaBackTester
from .order import Order


logger = getLogger(__name__)


class Positions:
    """Get Open Positions.

    Attributes:
        mt5 (MetaTrader): MetaTrader instance.
    """

    mt5: MetaTrader | MetaBackTester
    positions: tuple[TradePosition, ...]
    total_positions: int

    def __init__(self):
        """Get Open Positions"""
        self.config = Config()
        self.mt5 = MetaTrader() if self.config.mode != "backtest" else MetaBackTester()
        self.positions = ()
        self.total_positions = 0

    @backoff_decorator
    async def get_positions(self, *, symbol: str = None, ticket: int = None, group: str = None) -> tuple[TradePosition, ...]:
        """Get open positions with the ability to filter by symbol, ticket or group of symbols.
        Args:
            symbol (Optional[str]): Financial instrument name. If symbol is provided, ticket is ignored.
            ticket (Optional[int]): Position ticket.
            group (Optional[str]): Group of symbols.

        Returns:
            tuple[TradePosition, ...]: A tuple of open trade positions
        """
        kwargs = {}
        if symbol is not None:
            kwargs["symbol"] = symbol
            ticket = None
        if ticket is not None:
            kwargs["ticket"] = ticket
        if group is not None:
            kwargs["group"] = group
        positions = await self.mt5.positions_get(**kwargs)
        if positions is not None:
            self.positions = tuple(TradePosition(**pos._asdict()) for pos in positions)
            self.total_positions = len(self.positions)
            return self.positions
        logger.warning("Failed to get open positions")
        return ()

    async def get_position_by_ticket(self, *, ticket: int) -> TradePosition | None:
        """Get an open position by ticket.
        Args:
            ticket (int): Position ticket.

        Returns:
            TradePosition: Return an open position
        """
        positions = await self.mt5.positions_get(ticket=ticket)
        position = positions[0] if positions else None
        if position is None or position.ticket != ticket:
            return None
        return TradePosition(**position._asdict())

    async def get_positions_by_symbol(self, *, symbol: str) -> tuple[TradePosition, ...]:
        """Get open positions by symbol.
        Args:
            symbol (str): Financial instrument name.

        Returns:
            tuple[TradePosition, ...]: A tuple of open trade positions
        """
        positions = await self.mt5.positions_get(symbol=symbol)
        return tuple(TradePosition(**pos._asdict()) for pos in (positions or ()))

    @staticmethod
    async def close(*, ticket: int, symbol: str, price: float, volume: float, order_type: OrderType) -> OrderSendResult:
        """Close an open position for the trading account using the ticket and other parameters.

        Args:
            ticket (int): Position ticket.
            symbol (str): Financial instrument name.
            price (float): Closing price.
            volume (float): Volume to close.
            order_type (OrderType): Order type.
        """
        order = Order(
            action=TradeAction.DEAL,
            price=price,
            position=ticket,
            symbol=symbol,
            volume=volume,
            type=order_type.opposite,
        )
        return await order.send()

    async def close_position_by_ticket(self, *, ticket: int) -> OrderSendResult | None:
        """Close an open position using the ticket."""
        position = await self.get_position_by_ticket(ticket=ticket)
        if position is None:
            return None
        order = Order(
            position=position.ticket,
            symbol=position.symbol,
            volume=position.volume,
            type=position.type.opposite,
            price=position.price_current,
            action=TradeAction.DEAL,
        )
        return await order.send()

    @staticmethod
    async def close_position(*, position: TradePosition):
        """Close an open position for the trading account. Using a position object."""
        order = Order(
            position=position.ticket,
            symbol=position.symbol,
            volume=position.volume,
            type=position.type.opposite,
            price=position.price_current,
            action=TradeAction.DEAL,
        )
        return await order.send()

    async def close_all(self) -> int:
        """Close all open positions for the trading account. Specify a symbol or group to filter positions.

        Returns:
            int: Return number of positions closed.
        """
        positions = self.positions or await self.get_positions()
        results = await asyncio.gather(
            *(self.close_position(position=position) for position in positions), return_exceptions=True
        )
        return len([res for res in results if (isinstance(res, OrderSendResult) and res.retcode == 10009)])

    async def get_total_positions(self) -> int:
        """Get total number of open positions."""
        total = await self.mt5.positions_total()
        self.total_positions = total or self.total_positions
        return self.total_positions



import asyncio
import logging
from typing import Type, Iterable, Callable, Coroutine
from datetime import datetime, UTC

from MetaTrader5 import Tick

from .executor import Executor
from ..core.config import Config
from ..core.backtesting.backtest_controller import BackTestController
from ..core.meta_backtester import MetaBackTester
from ..core.backtesting.backtest_engine import BackTestEngine
from ..core.constants import CopyTicks
from .symbol import Symbol as Symbol
from .strategy import Strategy as Strategy

logger = logging.getLogger(__name__)


class BackTester:
    """The bot class. Create a bot instance to run your strategies.

    Attributes:
        executor: The default thread executor.
        config (Config): Config instance
        mt (MetaBackTester): MetaTrader instance
    """

    config: Config
    executor: Executor
    mt: MetaBackTester
    backtest_engine: BackTestEngine
    backtest_controller: BackTestController
    strategies: list[Strategy]

    def __init__(self, *, backtest_engine: BackTestEngine):
        self.config = Config()
        self.executor = Executor()
        self.mt = MetaBackTester()
        self.backtest_engine = backtest_engine
        self.backtest_controller = BackTestController()
        self.strategies = []

    def initialize_sync(self):
        """Prepares the bot by signing in to the trading account and initializing the symbols for the trading session.
        Starts the global task queue.

        Raises:
            SystemExit if sign in was not successful
        """
        try:
            self.mt.initialize_sync()
            login = self.mt.login_sync()
            if not login:
                logger.critical(f"Unable to sign in to MetaTrder 5 Terminal")
                raise Exception("Unable to sign in to MetaTrader 5 Terminal")
            logger.info("Login Successful")
            self.backtest_engine.setup_account_sync()
            self.init_strategies_sync()
            if (strategies := len(self.executor.strategy_runners)) == 0:
                logger.warning("No strategies were added to the backtester. Exiting ...")
                raise Exception("No strategies added to the backtester")
            self.config.task_queue.worker_timeout = 5
            self.add_coroutine(coroutine=self.config.task_queue.run, on_separate_thread=True)
            self.add_coroutine(coroutine=self.executor.exit)
            self.add_coroutine(coroutine=self.backtest_controller.control, on_separate_thread=True)
            parties = strategies + 1
            self.backtest_controller.set_parties(parties=parties)
        except Exception as err:
            logger.error(f"{err}. Backtester initialization failed")
            raise SystemExit

    async def initialize(self):
        """Prepares the bot by signing in to the trading account and initializing the symbols for the trading session.
        Starts the global task queue.

        Raises:
            SystemExit if sign in was not successful
        """
        try:
            await self.mt.initialize()
            login = await self.mt.login()
            if not login:
                logger.critical(f"Unable to sign in to MetaTrder 5 Terminal")
                raise Exception("Unable to sign in to MetaTrader 5 Terminal")
            logger.info("Login Successful")
            await self.backtest_engine.setup_account()
            await self.init_strategies()
            if (strategies := len(self.executor.strategy_runners)) == 0:
                logger.warning("No strategies were added to the backtester. Exiting ...")
                raise Exception("No strategies added to the backtester")
            self.config.task_queue.worker_timeout = 5
            self.add_coroutine(coroutine=self.config.task_queue.run, on_separate_thread=True)
            self.add_coroutine(coroutine=self.executor.exit)
            self.add_coroutine(coroutine=self.backtest_controller.control, on_separate_thread=True)
            parties = strategies + 1
            self.backtest_controller.set_parties(parties=parties)
        except Exception as err:
            logger.error(f"{err}. Backtester initialization failed")
            raise SystemExit

    def add_coroutine(self, *, coroutine: Callable[..., ...] | Coroutine, on_separate_thread=False, **kwargs):
        """Add a coroutine to the executor.

        Args:
            coroutine (Coroutine): A coroutine to be executed
            on_separate_thread (bool): Run the coroutine
            **kwargs (dict): keyword arguments for the coroutine

        Returns:

        """
        self.executor.add_coroutine(coroutine=coroutine, kwargs=kwargs, on_separate_thread=on_separate_thread)

    def execute(self):
        """Execute the bot."""
        self.initialize_sync()
        self.executor.execute()

    async def start(self):
        """Initialize the bot and execute it. Similar to calling `execute` method but is a coroutine."""
        await self.initialize()
        self.executor.execute()

    def add_strategy(self, *, strategy: Strategy):
        """Add a strategy to the list of strategies.
        An added strategy will only run if it's symbol was successfully initialized and it is added to the executor.

        Args:
            strategy (Strategy): A Strategy instance to run on bot

        Notes:
            Make sure the symbol has been added to the market
        """
        self.strategies.append(strategy)

    def add_strategies(self, *, strategies: Iterable[Strategy]):
        """Add multiple strategies at the same time

        Args:
            strategies: A list of strategies
        """
        [self.add_strategy(strategy=strategy) for strategy in strategies]

    def add_strategy_all(
        self, *, strategy: Type[Strategy], params: dict | None = None, symbols: list[Symbol] = None, **kwargs
    ):
        """Use this to run a single strategy on multiple symbols with the same parameters and keyword arguments.

        Keyword Args:
            strategy (Strategy): Strategy class
            params (dict): A dictionary of parameters for the strategy
            symbols (list): A list of symbols to run the strategy on
            **kwargs: Additional keyword arguments for the strategy
        """
        [self.add_strategy(strategy=strategy(symbol=symbol, params=params, **kwargs)) for symbol in symbols]

    async def init_strategy(self, *, strategy: Strategy) -> bool:
        """Initialize a single strategy. This method is called internally by the bot."""
        res = await strategy.symbol.initialize()
        if res:
            self.executor.add_strategy(strategy=strategy)
        return res

    def init_strategy_sync(self, *, strategy: Strategy) -> bool:
        """Initialize a single strategy. This method is called internally by the bot."""
        try:
            if self.backtest_engine.use_terminal is False:
                info = self.backtest_engine.symbols[strategy.symbol.name]
                tick = self.backtest_engine.prices[strategy.symbol.name].loc[self.backtest_engine.cursor.time]
                tick = Tick(tick)
                info = info._asdict() | {
                    "bid": tick.bid,
                    "bidhigh": tick.bid,
                    "bidlow": tick.bid,
                    "ask": tick.ask,
                    "askhigh": tick.ask,
                    "asklow": tick.bid,
                    "last": tick.last,
                    "volume_real": tick.volume_real,
                }
                strategy.symbol.set_attributes(**info)
                self.executor.add_strategy(strategy=strategy)
                return True
            else:
                self.mt._symbol_select(strategy.symbol.name, True)
                self.mt._market_book_add(strategy.symbol.name)
                info = self.mt._symbol_info(strategy.symbol.name)
                time = datetime.fromtimestamp(self.backtest_engine.cursor.time, tz=UTC)
                tick = self.mt._copy_ticks_from(strategy.symbol.name, time, 1, CopyTicks.ALL)
                tick = Tick(tick[-1]) if tick is not None else None
                info = info._asdict() | {
                    "bid": tick.bid,
                    "bidhigh": tick.bid,
                    "bidlow": tick.bid,
                    "ask": tick.ask,
                    "askhigh": tick.ask,
                    "asklow": tick.bid,
                    "last": tick.last,
                    "volume_real": tick.volume_real,
                }
                strategy.symbol.set_attributes(**info)
                self.executor.add_strategy(strategy=strategy)
                return True
        except Exception as err:
            logger.error("%s: Unable to initialize strategy", err)
            return False

    async def init_strategies(self):
        """Initialize the symbols for the current trading session. This method is called internally by the bot."""
        tasks = [self.init_strategy(strategy=strategy) for strategy in self.strategies]
        await asyncio.gather(*tasks)

    def init_strategies_sync(self):
        """Initialize the symbols for the current trading session. This method is called internally by the bot."""
        [self.init_strategy_sync(strategy=strategy) for strategy in self.strategies]



from logging import getLogger
from typing import Self

from ..core.base import _Base
from ..core.models import AccountInfo
from ..core.exceptions import LoginError

logger = getLogger(__name__)


class Account(_Base, AccountInfo):
    """A singleton class for managing a trading account. A subclass of _Base and AccountInfo. It supports
    Asynchronous context management protocol.

    Attributes:
        connected (bool): Status of connection to MetaTrader 5 Terminal
    """

    _instance: Self
    connected: bool

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "_instance"):
            cls._instance = super().__new__(cls)
            cls._instance.connected = False
        return cls._instance

    async def __aenter__(self) -> Self:
        """Connect to a trading account and return the account instance.
        Async context manager for the Account class.

        Returns:
            Account: An instance of the Account class

        Raises:
            LoginError: If login fails
        """
        await self.mt5.initialize()
        self.connected = await self.mt5.login()
        if not self.connected:
            raise LoginError(f"Login failed: {self.mt5.error}")
        await self.refresh()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.mt5.shutdown()
        self.connected = False

    async def refresh(self):
        """Refreshes the account instance with the latest account details from the MetaTrader 5 terminal"""
        account_info = await self.mt5.account_info()
        acc = account_info._asdict()
        self.connected = True
        self.set_attributes(**acc)



"""This module contains the Records class, which is used to read and update trade records from csv files."""

import asyncio
import json
from pathlib import Path
import csv
import logging
from typing import Iterable

from MetaTrader5 import TradePosition

from ..core.config import Config
from ..core.meta_trader import MetaTrader
from ..core.meta_backtester import MetaBackTester

logger = logging.getLogger(__name__)


class TradeRecords:
    """This utility class read trade records from csv files, and update them based on their closing positions.

    Attributes:
        config: Config object
        records_dir(Path): Absolute path to directory containing record of placed trades, If not given takes the default
        from the config
    """

    config: Config
    mt5: MetaTrader | MetaBackTester
    positions: list[TradePosition] | None = None

    def __init__(self, *, records_dir: Path | str = ""):
        """Initialize the Records class. The main method of this class is update_records which you should call to update
        all the records specified in the records_dir.

        Keyword Args:
            records_dir (Path): Absolute path to directory containing record of placed trades.
        """
        self.config = Config()
        self.mt5 = MetaTrader() if self.config.mode != "backtest" else MetaBackTester()
        self.records_dir = records_dir or self.config.records_dir

    def get_csv_records(self):
        """Get trade records saved as csv from records_dir folder

        Yields:
            files: Trade record files
        """
        for file in self.records_dir.iterdir():
            if file.is_file() and file.name.endswith(".csv"):
                yield file

    def get_json_records(self):
        """Get trade records from records_dir folder

        Yields:
            files (Path): Trade record files
        """
        for file in self.records_dir.iterdir():
            if file.is_file() and file.name.endswith(".json"):
                yield file

    async def read_update_csv(self, *, file: Path):
        """Read and update csv trade records

        Args:
            file: Trade record file in csv format
        """
        try:
            with open(file, mode="r", newline="") as fr:
                reader: Iterable[dict] | csv.DictReader = csv.DictReader(fr)
                rows = [row for row in reader]
                rows = await self.update_rows(rows=rows)

            with open(file, mode="w", newline="") as fw:
                writer = csv.DictWriter(fw, fieldnames=reader.fieldnames, extrasaction="ignore", restval=None)
                writer.writeheader()
                writer.writerows(rows)
        except Exception as err:
            logger.error(f"Error: {err}. Unable to read and update csv trade records")

    async def read_update_json(self, *, file: Path):
        """Read and update json trade records
        Args:
            file: Trade record file in csv format
        """
        try:
            with open(file, mode="r") as fh:
                data = json.load(fh)
                rows = [row for row in data]
                rows = await self.update_rows(rows=rows)

            with open(file, mode="w") as fh:
                json.dump(rows, fh, indent=2)
        except Exception as err:
            logger.error(f"Error: {err}. Unable to read and update json trade records")

    async def update_row(self, *, row: dict) -> dict:
        """Update a single row of entered trade in the csv or json file with the actual profit.

        Args:
            row: A dictionary from the dictionary writer object of the csv file.

        Returns:
            dict: A dictionary with the actual profit and win status.
        """
        try:
            order = int(row["order"])
            positions = self.positions or await self.mt5.positions_get()
            position_ids = [position.ticket for position in positions]
            deals = await self.mt5.history_deals_get(position=order)
            if not deals or len(deals) <= 1:
                return row
            deals = [
                deal
                for deal in deals
                if (
                    deal.order != deal.position_id
                    and deal.position_id == order
                    and deal.entry == 1
                    and deal.position_id not in position_ids
                )
            ]
            deals.sort(key=lambda deal: deal.time_msc)
            deal = deals[-1]
            row.update(actual_profit=deal.profit, win=deal.profit > 0, closed=True)
            return row
        except Exception as err:
            logging.error(f"Error: {err}. Unable to update trade record")
            return row

    async def update_rows(self, *, rows: list[dict]) -> list[dict]:
        """Update the rows of entered trades in the csv or json file with the actual profit.

        Args:
            rows: A list of dictionaries.

        Returns:
            list[dict]: A list of dictionaries with the actual profit and win status.
        """
        self.positions = await self.mt5.positions_get()
        closed, unclosed = [], []
        for row in rows:
            closed_ = row.get("closed", False)
            closed_ = closed_.title() == "True" if isinstance(closed_, str) else closed_
            if closed_:
                closed.append(row)
            else:
                unclosed.append(row)
        unclosed = await asyncio.gather(*[self.update_row(row=row) for row in unclosed])
        return closed + list(unclosed)

    async def update_csv_records(self):
        """Update csv trade records in the records_dir folder."""
        records = [self.read_update_csv(file=record) for record in self.get_csv_records()]
        await asyncio.gather(*records)

    async def update_json_records(self):
        """Update json trade records in the records_dir folder."""
        records = [self.read_update_json(file=record) for record in self.get_json_records()]
        await asyncio.gather(*records)

    async def update_csv_record(self, *, file: Path | str):
        """Update a single trade record csv file."""
        await self.read_update_csv(file=file)

    async def update_json_record(self, *, file: Path | str):
        """Update a single json trade record file"""
        await self.read_update_json(file=file)



"""Utility functions for aiomql."""

import decimal
import random
from functools import wraps, partial
import asyncio
from threading import RLock
from logging import getLogger
from .core.config import Config

logger = getLogger(__name__)


async def backtest_sleep(secs: float):
    config = Config()
    btc = config.backtest_controller
    try:
        if btc.parties == 2:
            steps = int(secs) // config.backtest_engine.speed
            steps = max(steps, 1)
            config.backtest_engine.fast_forward(steps=steps)
            btc.wait()

        elif btc.parties > 2:
            _time = config.backtest_engine.cursor.time + secs
            while _time > config.backtest_engine.cursor.time:
                btc.wait()
        else:
            btc.wait()
    except Exception as err:
        btc.wait()
        logger.error("Error: %s in backtest_sleep", err)


# async def backtest_sleep(secs):
#     """An async sleep function for use during backtesting."""
#     btc = BackTestController()
#     config = Config()
#     sleep = config.backtest_engine.cursor.time + secs
#     while sleep > config.backtest_engine.cursor.time:
#         btc.wait()


def dict_to_string(data: dict, multi=False) -> str:
    """Convert a dict to a string. Useful for logging.

    Args:
        data (dict): The dict to convert.
        multi (bool, optional): If True, each key-value pair will be on a new line. Defaults to False.

    Returns:
        str: The string representation of the dict.
    """
    sep = "\n" if multi else ", "
    return f"{sep}".join(f"{key}: {value}" for key, value in data.items())


def backoff_decorator(func=None, *, max_retries: int = 2, retries: int = 0, error="") -> callable:
    """A decorator to retry a function a number of times before giving up.
    Args:
        func (callable, optional): The function to decorate. Defaults to None.
        max_retries (int, optional): The maximum number of retries. Defaults to 2.
        retries (int, optional): The number of retries. Defaults to 0.
        error (Any, optional): The error to raise when the maximum number of retries is reached. Defaults to "".
    """
    if func is None:
        return partial(backoff_decorator, max_retries=max_retries, retries=retries, error=error)

    @wraps(func)
    async def wrapper(*args, **kwargs):
        nonlocal retries
        if max_retries == retries:
            retries = 0
            await func(*args, **kwargs)
        try:
            retries += 1
            res = await func(*args, **kwargs)
            if error != "" and res == error:
                raise TypeError("Invalid return type")
            else:
                retries = 0
                return res
        except Exception as err:
            logger.error("Error in %s: %s", func.__name__, err)
            if Config().mode != "backtest":
                await asyncio.sleep(2**retries + random.uniform(0, max_retries))
            await wrapper(*args, **kwargs)

    return wrapper


def error_handler(func=None, *, msg="", exe=Exception, response=None, log_error_msg=True):
    """A decorator to handle errors in an async function.
    Args:
        func (callable, optional): The function to decorate. Defaults to None.
        msg (str, optional): The error message to log. Defaults to "".
        exe (Exception, optional): The exception to catch. Defaults to Exception.
        response (Any, optional): The response to return when an error occurs. Defaults to None.
        log_error_msg (bool, optional): If True, log the error message. Defaults to True.
    """
    if func is None:
        return partial(error_handler, msg=msg, exe=exe, response=response, log_error_msg=log_error_msg)

    @wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            res = await func(*args, **kwargs)
            return res
        except exe as err:
            if log_error_msg:
                logger.error(f"Error in {func.__name__}: {msg or err}")
            return response

    return wrapper


def error_handler_sync(func=None, *, msg="", exe=Exception, response=None, log_error_msg=True):
    """A decorator to handle errors in a synchronous function.

    Args:
        func (callable, optional): The function to decorate. Defaults to None.
        msg (str, optional): The error message to log. Defaults to "".
        exe (Exception, optional): The exception to catch. Defaults to Exception.
        response (Any, optional): The response to return when an error occurs. Defaults to None.
        log_error_msg (bool, optional): If True, log the error message. Defaults to True.
    """
    if func is None:
        return partial(error_handler, msg=msg, exe=exe, response=response, log_error_msg=log_error_msg)

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            res = func(*args, **kwargs)
            return res
        except exe as err:
            if log_error_msg:
                logger.error(f"Error in {func.__name__}: {msg or err}")
            return response

    return wrapper


def round_down(value: int | float, base: int) -> int | float:
    """Round down a number to the nearest base.
    Args:
        value (int | float): The number to round down.
        base (int): The base to round down to.

    Returns:
        (int | float): The rounded down number.
    """
    return int(value) if value % base == 0 else int(value - (value % base))


def round_up(value: int | float, base: int) -> int:
    return int(value) if value % base == 0 else int(value + base - (value % base))


# noinspection PyShadowingNames
def round_off(value: float, step: float, round_down: bool = False) -> float:
    """Round off a number to the nearest step.

    Args:
        value (float): The number to round off.
        step (float): The step to round off to.
        round_down (bool, optional): If True, the number is rounded down otherwise it is rounded up. Defaults to False.

    Returns:
        float: The rounded off number.
    """
    with decimal.localcontext() as ctx:
        ctx.rounding = decimal.ROUND_DOWN if round_down else decimal.ROUND_UP
        return float(decimal.Decimal(str(value)).quantize(decimal.Decimal(str(step))))


def async_cache(fun):
    """A decorator to cache the result of an async function."""
    @wraps(fun)
    async def wrapper(*args, **kwargs):
        key = (args, frozenset(kwargs.items()))
        with wrapper.lock:
            if key not in wrapper.cache:
                wrapper.cache[key] = await fun(*args, **kwargs)
            return wrapper.cache[key]

    wrapper.lock = RLock()
    wrapper.cache = {}
    return wrapper



from abc import ABC, abstractmethod
from datetime import datetime, UTC
from typing import TypeVar
from logging import getLogger


from ..core.models import OrderType, OrderSendResult, OrderCheckResult
from ..core.config import Config
from ..core.task_queue import QueueItem
from .result import Result
from .order import Order
from .symbol import Symbol as _Symbol
from .ram import RAM
from .._utils import error_handler

logger = getLogger(__name__)
Symbol = TypeVar("Symbol", bound=_Symbol)


class Trader(ABC):
    """Base class for creating and managing orders.
    Handles the creation and placing of an order.
    It is an abstract class and must be subclassed to implement the place_trade method.

    Attributes:
        symbol (Symbol): The financial instrument.
        ram (RAM): RAM instance
        order (Order): Trade order
        parameters (dict): Parameters of the trading strategy used to place the trade
        config (Config): The Config instance.
    """

    config: Config
    ram: RAM
    parameters: dict

    def __init__(self, *, symbol: Symbol, ram: RAM = None):
        """Initializes the order object and RAM instance

        Args:
            symbol (Symbol): Financial instrument
            ram (RAM): Risk Assessment and Management instance
        """
        self.config = Config()
        self.symbol = symbol
        self.ram = ram or RAM()
        self.order = Order(symbol=symbol.name)
        self.parameters = {}

    def set_trade_stop_levels_pips(self, *, pips: float, risk_to_reward: float = None):
        """Sets the stop loss and take profit for the order.
        This method uses pips as defined for forex instruments. It is assumed that order_type and price are already
        set before calling this method.

        Args:
            pips (float): Target pips
            risk_to_reward (float): Optional risk to reward ratio
        """
        pips = pips * self.symbol.pip
        sl, tp = pips, pips * (risk_to_reward or self.ram.risk_to_reward)
        price = self.order.price
        if self.order.type == OrderType.BUY:
            self.order.sl, self.order.tp = round(price - sl, self.symbol.digits), round(price + tp, self.symbol.digits)
        elif self.order.type == OrderType.SELL:
            self.order.sl, self.order.tp = round(price + sl, self.symbol.digits), round(price - tp, self.symbol.digits)

    def set_trade_stop_levels_points(self, *, points: float, risk_to_reward: float = None):
        """Set the stop loss and take profit levels of the order based on the points and the risk to reward ratio.
           It is assumed that order_type and price are already set before calling this method.

        Args:
            points (float): Target points
            risk_to_reward (float): Risk to reward ratio
        """
        points = points * self.symbol.point
        sl, tp = points, points * (risk_to_reward or self.ram.risk_to_reward)
        price, digits = self.order.price, self.symbol.digits

        if self.order.type == OrderType.BUY:
            self.order.sl, self.order.tp = round(price - sl, self.symbol.digits), round(price + tp, digits)

        elif self.order.type == OrderType.SELL:
            self.order.sl, self.order.tp = round(price + sl, self.symbol.digits), round(price - tp, digits)

    async def create_order_with_stops(
        self, *, order_type: OrderType, sl: float, tp: float, amount_to_risk: float = None
    ):
        """Create an order with stop loss and take profit levels. Use the amount to risk per trade to
        calculate the volume.

        Args:
            order_type (OrderType): Order type
            sl (float): Stop loss in price
            tp (float): Take profit in price
            amount_to_risk (float): Amount to risk per trade in terms of the account currency. Optional parameter,
                default is the amount as computed by the RAM instance.
        """
        amount = amount_to_risk or await self.ram.get_amount()
        amount = await self.symbol.amount_in_quote_currency(amount=amount)
        tick = await self.symbol.info_tick()
        price = tick.ask if order_type == OrderType.BUY else tick.bid
        volume = await self.symbol.compute_volume_sl(amount=amount, price=price, sl=sl)
        self.order.set_attributes(sl=sl, tp=tp, volume=volume, price=price, type=order_type)

    async def create_order_with_sl(
        self, *, order_type: OrderType, sl: float, amount_to_risk: float = None, risk_to_reward: float = None
    ):
        """
        Create an order with a given stop_loss level. Use the amount to risk per trade to calculate the volume.

        Args:
            order_type (OrderType): Order type
            sl (float): Stop loss in price
            amount_to_risk (float): Amount to risk per trade in terms of the account currency. Optional parameter,
                default is the amount as computed by the RAM instance.
            risk_to_reward (float): Risk to reward ratio. Optional parameter, default is the risk to reward ratio as
                defined in the RAM instance.
        """
        amount = amount_to_risk or await self.ram.get_amount()
        amount = await self.symbol.amount_in_quote_currency(amount=amount)
        tick = await self.symbol.info_tick()
        price = tick.ask if order_type == OrderType.BUY else tick.bid
        dsl = abs(price - sl)
        dtp = dsl * (risk_to_reward or self.ram.risk_to_reward)
        tp = price + dtp if order_type == OrderType.BUY else price - dtp
        volume = await self.symbol.compute_volume_sl(amount=amount, price=price, sl=sl)
        self.order.set_attributes(sl=sl, tp=tp, volume=volume, price=price, type=order_type)

    async def create_order_with_points(
        self, *, order_type: OrderType, points: float, amount_to_risk: float = None, risk_to_reward: float = None
    ):
        """Create an order with specific points to risk. Use the amount to risk per trade to calculate the volume.

        Args:
            order_type (OrderType): Order type
            points (float): Points to risk
            amount_to_risk (float): Amount to risk per trade in terms of the account currency. Optional parameter,
                default is the amount as computed by the RAM instance.
            risk_to_reward (float): Risk to reward ratio. Optional parameter, default is the risk to reward ratio as
                defined in the RAM instance.
        """
        self.order.type = order_type
        amount = amount_to_risk or await self.ram.get_amount()
        amount = await self.symbol.amount_in_quote_currency(amount=amount)
        tick = await self.symbol.info_tick()
        self.order.price = tick.ask if order_type == OrderType.BUY else tick.bid
        volume = await self.symbol.compute_volume_points(amount=amount, points=points)
        self.order.volume = volume
        self.set_trade_stop_levels_points(points=points, risk_to_reward=risk_to_reward)

    async def create_order_no_stops(self, *, order_type: OrderType, volume: float = None):
        """Create an order without setting stop loss and take profit. Using minimum lot size.

        Args:
            order_type (OrderType): Order type
            volume (float): Volume to trade with. Optional parameter, default is the minimum lot size.
        """
        tick = await self.symbol.info_tick()
        self.order.volume = volume or self.symbol.volume_min
        self.order.price = tick.ask if order_type == OrderType.BUY else tick.bid
        self.order.type = order_type

    async def check_order(self) -> OrderCheckResult | None:
        """Check order before sending it to the broker.

        Returns:
            bool: True if order can go through else false
        """
        check = await self.order.check()

        if check is None:
            logger.warning(f"{self.order.mt5.error}: Order check failed")
            return check

        if check.retcode != 0:
            logger.warning("Invalid order %s, for due to %s", self.symbol, check.comment)
        return check

    async def send_order(self) -> OrderSendResult | None:
        """Send the order to the broker."""
        result = await self.order.send()
        if result is None:
            logger.warning("%s: Failed to place order.", self.order.mt5.error)
            return result

        if result.retcode != 10009:
            logger.warning("Unable to place order for %s due to %s", self.symbol, result.comment)
            return result
        return result

    @error_handler
    async def record_trade(self, *, result: OrderSendResult, parameters: dict = None, name: str = ""):
        """Record the trade in csv or json.
        Args:
            result (OrderSendResult): Result of the order send
            parameters (dict): parameters of the trading strategy used to place the trade
            name (str): Name of the trading strategy
        """
        if self.config.record_trades is False or result.retcode != 10009:
            return
        params = {**parameters} if isinstance(parameters, dict) else {}
        profit = await self.order.calc_profit()
        params["expected_profit"] = profit
        date = (
            datetime.now(tz=UTC)
            if self.config.mode == "live"
            else datetime.fromtimestamp(self.config.backtest_engine.cursor.time, tz=UTC)
        )
        params["date"] = date.strftime("%Y-%m-%d %H:%M:%S.%f")
        res = Result(result=result, parameters=params, name=name)
        self.config.task_queue.add(item=QueueItem(res.save), must_complete=True)

    @abstractmethod
    async def place_trade(self, *args, **kwargs):
        """Places a trade based on the order_type."""



import csv
import json
from logging import getLogger
from typing import Iterable, Literal
from threading import Lock

from ..core.config import Config
from ..core.models import OrderSendResult

logger = getLogger(__name__)


class Result:
    """A base class for handling trade results and strategy parameters for record keeping and reference purpose.

    Attributes:
        config (Config): The configuration object
        name: Any desired name for the result file object
    """

    config: Config
    lock = Lock()

    def __init__(self, *, result: OrderSendResult, parameters: dict = None, name: str = ""):
        """
        Prepare result data
        Args:
            result:
            parameters:
            name:
        """
        self.config = Config()
        self.parameters = parameters or {}
        self.result = result
        self.name = name or self.parameters.get("name", "Trades")

    def get_data(self) -> dict:
        res = self.result.get_dict(exclude={"retcode", "comment", "retcode_external", "request_id", "request"})
        return self.parameters | res | {"actual_profit": 0, "closed": False, "win": False}

    async def save(self, *, trade_record_mode: Literal["csv", "json"] = None):
        """Record trade results as a csv or json file
        Args:
            trade_record_mode (Literal['csv'|'json']): Mode of saving trade records
        """
        with self.lock:
            trade_record_mode = trade_record_mode or self.config.trade_record_mode
            if trade_record_mode == "csv":
                await self.to_csv()
            elif trade_record_mode == "json":
                await self.to_json()
            else:
                logger.error(f"Invalid trade record mode: {trade_record_mode}")

    async def to_csv(self):
        """Record trade results and associated parameters as a csv file"""
        try:
            data = self.get_data()
            file = self.config.records_dir / f"{self.name}.csv"
            file.touch(exist_ok=True) if not file.exists() else ...
            read_file = file.open("r", newline="")
            reader: Iterable[dict] = csv.DictReader(read_file)
            rows: list[dict] = []
            headers = set()
            [(rows.append(row), headers.update(row.keys())) for row in reader]
            rows.append(data)
            headers.update(data.keys())
            read_file.close()
            with file.open("w", newline="") as write_file:
                writer = csv.DictWriter(write_file, fieldnames=headers, restval=None, extrasaction="ignore")
                writer.writeheader()
                writer.writerows(rows)
        except Exception as err:
            logger.error(f"Unable to save to csv: {err}")

    @staticmethod
    def serialize(value) -> str:
        """Serialize the trade records and strategy parameters"""
        try:
            return str(value)
        except Exception as err:
            logger.error("%s: Unable to serialize value", err)
            return ""

    async def to_json(self):
        """Save trades and strategy parameters in a json file"""
        try:
            file = self.config.records_dir / f"{self.name}.json"
            data = self.get_data()
            if not file.exists():
                file.touch()
                with file.open("w") as fh:
                    json.dump([], fh, indent=2)

            with file.open("r") as fh:
                rows = json.load(fh)
                rows.append(data)

            with file.open("w") as fh:
                json.dump(rows, fh, indent=2, skipkeys=True, default=self.serialize)

        except Exception as err:
            logger.error(f"Unable to save as json file: {err}")



"""Exceptions for the aiomql package."""

__all__ = ["LoginError", "VolumeError", "SymbolError", "OrderError"]


class LoginError(Exception):
    """Raised when an error occurs when logging in."""

    ...


class VolumeError(Exception):
    """Raised when a volume is not valid or out of range for a symbol."""

    ...


class SymbolError(Exception):
    """Raised when a symbol is not provided where required or not available in the Market Watch."""

    ...


class OrderError(Exception):
    """Raised when an error occurs when working with the order class."""

    ...


class StopTrading(Exception):
    """Raised when the user wants to stop trading."""

    ...



from threading import RLock
import asyncio
import json
from datetime import datetime, UTC
from typing import Literal
import random
from functools import cached_property
from logging import getLogger
from pathlib import Path

import pandas as pd
import numpy as np
from pandas import DataFrame
from MetaTrader5 import (
    Tick,
    SymbolInfo,
    AccountInfo,
    TradeOrder,
    TradePosition,
    TradeDeal,
    TradeRequest,
    OrderCheckResult,
    OrderSendResult,
    TerminalInfo,
)

from ..meta_trader import MetaTrader
from ..constants import (
    TimeFrame,
    OrderType,
    TradeAction,
    AccountStopOutMode,
    PositionReason,
    DealType,
    DealReason,
    DealEntry,
    OrderReason,
    CopyTicks,
)

from ..._utils import round_down, round_up, error_handler, error_handler_sync, async_cache

from .get_data import BackTestData, GetData, Cursor
from .backtest_account import BackTestAccount
from .trades_manager import PositionsManager, OrdersManager, DealsManager

logger = getLogger(__name__)


# noinspection PyUnresolvedReferences
class BackTestEngine:
    mt5: MetaTrader
    span: range
    range: range
    speed: int
    cursor: Cursor
    iter: zip
    rates: dict[str, dict[int, DataFrame]]
    ticks: dict[str, DataFrame]
    prices: dict[str, DataFrame]
    orders: OrdersManager
    deals: DealsManager
    positions: PositionsManager
    _account: BackTestAccount
    stop_testing: bool
    use_terminal: bool
    restart: bool
    stop_time: int | None
    close_open_positions_on_exit: bool
    preloaded_ticks: dict[str, DataFrame]
    preload: bool
    account_lock: RLock
    account_info: dict

    def __init__(
        self,
        *,
        data: BackTestData = None,
        speed: int = 60,
        start: float | datetime = 0,
        end: float | datetime = 0,
        restart: bool = True,
        use_terminal: bool = None,
        name: str = "",
        stop_time: float | datetime = None,
        close_open_positions_on_exit: bool = True,
        preload=True,
        assign_to_config: bool = True,
        account_info: dict = None,
    ):
        self._data = data or BackTestData()
        self.mt5 = MetaTrader()
        self.config = self.mt5.config
        if assign_to_config:
            self.config.backtest_engine = self
        self.setup_test_range(start=start, end=end, speed=speed, restart=restart)
        self.setup_data(restart=restart)
        start, end = (
            (self.span[0], self.span[-1]) if len(self.span) >= 2 else ((now := datetime.now(UTC).timestamp()), now)
        )
        start, end = datetime.fromtimestamp(start, tz=UTC), datetime.fromtimestamp(end, tz=UTC)
        self.name = name or self._data.name or f"backtest_data_{start:%d_%m_%y}_{end:%d_%m_%y}"
        self.stop_testing = False
        self.use_terminal = self.config.use_terminal_for_backtesting if use_terminal is None else use_terminal
        self.close_open_positions_on_exit = close_open_positions_on_exit
        if stop_time is not None:
            val = (
                stop_time.astimezone(tz=UTC)
                if isinstance(stop_time, datetime)
                else datetime.fromtimestamp(stop_time, tz=UTC)
            )
            stop_time = int(val.timestamp())
        self.stop_time = stop_time
        self.preload = preload
        self.preloaded_ticks = {}
        self.account_lock = RLock()
        self.account_info = account_info or {}

    def __next__(self) -> Cursor:
        try:
            index, time = next(self.iter)
            if self.stop_time and time >= self.stop_time:
                raise StopIteration
            self.cursor = Cursor(index=index, time=time)
            return self.cursor
        except StopIteration:
            logger.critical("End of the test range")
            self.stop_testing = True

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def setup_test_range(
        self, *, start: float | datetime = None, end: float | datetime = None, speed: int = 60, restart: bool = True
    ):
        """Setup the test range for the backtest engine. This is used to set the range of the backtest and the speed
         at which it runs.

        Args:
            start (float | datetime, optional): The start time of the backtest. Defaults to None. If a float is passed,
             it is assumed to be a timestamp.

            end (float | datetime, optional): The end time of the backtest. Defaults to None. If a float is passed,
             it is assumed to be a timestamp.

            speed (int, optional): The speed of the backtest. Defaults to 60.

            restart (bool, optional): Whether to restart the backtest. Defaults to True.
             This is useful when resuming a backtest using a saved BackTestData.
        """
        if self._data.span and self._data.range:
            start = start or self._data.span[0]
            end = end or self._data.span[-1] + speed
        start = start.astimezone(tz=UTC) if isinstance(start, datetime) else datetime.fromtimestamp(start, tz=UTC)
        end = end.astimezone(tz=UTC) if isinstance(end, datetime) else datetime.fromtimestamp(end, tz=UTC)
        span_start = int(start.timestamp())
        span_end = int(end.timestamp())
        self.speed = speed
        self.span = range(span_start, span_end, speed)
        self.range = range(0, span_end - span_start, speed)
        self.iter = zip(self.range, self.span)

        if restart is False and self._data.cursor is not None:
            self.cursor = self._data.cursor
            index = self.cursor.index // speed
            new_range = range(self.range[index], self.range.stop, speed)
            new_span = range(self.span[index], self.span.stop, speed)
            self.iter = zip(new_range, new_span)
            # self.go_to(time=self.cursor.time)
        else:
            self.cursor = Cursor(index=self.range.start, time=self.span.start)

    def setup_data(self, *, restart: bool = True):
        """Sets up the data for the backtest engine. This includes the orders, positions, deals and account
         information. This data is handled by specialized classes such as the BackTestAccount and the TradeManager
         classes.

        Args:
            restart (bool, optional): Whether to restart the data. Defaults to True.
        """
        if restart is True:
            self.orders = OrdersManager()
            self.positions = PositionsManager()
            self.deals = DealsManager()
            self._account = BackTestAccount()
            return

        orders = {}
        for ticket, order in self._data.orders.items():
            orders[ticket] = TradeOrder((order.get(k) for k in TradeOrder.__match_args__))
        self.orders = OrdersManager(data=orders)

        positions = {}
        for ticket, position in self._data.positions.items():
            positions[ticket] = TradePosition((position.get(k) for k in TradePosition.__match_args__))
        self.positions = PositionsManager(
            data=positions, open_positions=self._data.open_positions, margins=self._data.margins
        )

        deals = {}
        for ticket, deal in self._data.deals.items():
            deals[ticket] = TradeDeal((deal.get(k) for k in TradeDeal.__match_args__))
        self.deals = DealsManager(data=deals)

        self._account = BackTestAccount(**self._data.account)

    def next(self) -> Cursor:
        """Move the cursor to the next time step in the backtest range."""
        return next(self)

    @property
    def data(self):
        """The BackTestData instance used for the backtest. If not provided, a new instance is created,
        and the data is made persistent when the backtest is stopped."""
        return self._data

    def reset(self, clear_data: bool = False):
        """Reset the backtest engine. This is useful when restarting the backtest from the beginning."""
        self.iter = zip(self.range, self.span)
        self.cursor = Cursor(index=self.range.start, time=self.span.start)
        if clear_data:
            self.setup_data(restart=True)

    def go_to(self, *, time: datetime | float):
        """Move the cursor to a specific time in the backtest range. You can pass a datetime object or a timestamp.
        You can't go back in time or beyond the limits of the range.
        """
        time = time.astimezone(tz=UTC) if isinstance(time, datetime) else datetime.fromtimestamp(time, tz=UTC)
        time = int(time.timestamp())
        steps = time - self.cursor.time
        steps = steps // self.speed
        steps = max(steps, 1)
        if 0 <= steps < (len(self.range) - 1):
            self.fast_forward(steps=steps)
            return
        raise ValueError("Can't go back in time or beyond the limits of the range")

    def fast_forward(self, *, steps: int):
        """Fast-forward the backtester by the given steps."""
        for _ in range(steps):
            self.next()

    @staticmethod
    def get_dtype(*, df: DataFrame) -> list[tuple[str, str]]:
        return [(c, t) for c, t in zip(df.columns, df.dtypes)]

    async def tracker(self):
        """The tracker monitors and updates open positions on every iteration. It is called by the controller."""
        try:
            pos_tasks = [self.check_position(ticket=ticket) for ticket in self.positions._open_positions]
            await asyncio.gather(*pos_tasks)
            profit = sum(pos.profit for pos in self.positions.open_positions)
            self.update_account(profit=profit)
            self.check_account()
        except Exception as exe:
            logger.critical("Error in tracker: %s at %d", exe, self.cursor.time)

    @error_handler_sync
    def save_result_to_json(self):
        """Saves the result to a json file at the end of testing."""
        data = self._account.get_dict(include={"balance", "profit", "equity", "margin", "margin_free", "margin_level"})
        wins = [position for ticket in self.positions if (position := self.positions.get(ticket)).profit > 0]
        losses = [position for ticket in self.positions if (position := self.positions.get(ticket)).profit <= 0]
        win = round(sum(position.profit for position in wins), self._account.currency_digits)
        loss = round(sum(position.profit for position in losses), self._account.currency_digits)
        profit_factor = round(abs(win / loss), 2) if loss != 0 else 0
        wins, losses, total = len(wins), len(losses), len(self.positions._data)
        win_percentage = round(wins / total * 100, 2) if total > 0 else 0
        net_profit = round(win - abs(loss), self._account.currency_digits)
        profitability = net_profit / (self._account.balance - net_profit) * 100 if net_profit != 0 else 0
        profitability = round(profitability, 2)
        data.update(
            {
                "wins": wins,
                "losses": losses,
                "total": total,
                "win_percentage": win_percentage,
                "win": win,
                "loss": loss,
                "net_profit": net_profit,
                "profit_factor": profit_factor,
                "profitability": profitability,
            }
        )
        path = Path(self.config.backtest_dir / f"{self.name}.json")
        with path.open("w") as file:
            json.dump(data, file, indent=4)

    async def close_all_open(self):
        """Closes all open position at the end of testing"""
        tasks = [self.check_position(ticket=position.ticket) for position in self.positions.open_positions]
        await asyncio.gather(*tasks)
        for position in self.positions.open_positions:
            await self.close_position_manually(ticket=position.ticket)

    @error_handler
    async def wrap_up(self):
        """Wraps up the backtest. This is called at the end of testing to save the results and close all open
        positions."""
        if self.close_open_positions_on_exit:
            await self.close_all_open()
        self.save_result_to_json()
        self._data.deals = self.deals.to_dict()
        self._data.orders = self.orders.to_dict()
        self._data.positions = self.positions.to_dict()
        self._data.open_positions = self.positions._open_positions
        self._data.margins = self.positions.margins
        self._data.account = self._account.asdict()
        self._data.cursor = self.cursor
        self._data.span = self.span
        self._data.range = self.range
        self._data.account = self._account.asdict()
        name = self._data.name or self.name
        self._data.name = name
        path = self.config.backtest_dir / f"{name}.pkl"
        GetData.pickle_data(data=self._data, name=path)

    async def preload_ticks(self, *, symbol: str):
        """Pull a month data on ticks from the terminal. Starting from the current time.

        Args:
            symbol (str): The symbol to preload ticks for.
        """
        try:
            start = self.cursor.time
            end = start + (30 * 24 * 60 * 60)
            end = end if end < self.span.stop else self.span.stop
            span = range(start, end)
            start = datetime.fromtimestamp(start, tz=UTC)
            end = datetime.fromtimestamp(end, tz=UTC)
            ticks = await self.mt5.copy_ticks_range(symbol, start, end, CopyTicks.ALL)
            ticks = pd.DataFrame(ticks)
            ticks.drop_duplicates(subset=["time"], keep="last", inplace=True)
            ticks.set_index("time", inplace=True, drop=False)
            ticks = ticks.reindex(span, method="nearest", copy=True)
            self.preloaded_ticks[symbol] = ticks
        except Exception as exe:
            logger.error(f"Error Preloading Ticks: {exe}")

    @async_cache
    async def get_price_tick(self, *, symbol: str, time: int) -> Tick | None:
        """Get the price tick for a symbol at a given time. If the preload option is set to True,
        it will use the preloaded ticks when available.

        Args:
            symbol (str): The symbol to get the price tick for.
            time (int): The time to get the price tick.
        """
        try:
            if self.use_terminal and self.preload:
                if (ticks := self.preloaded_ticks.get(symbol)) is not None and time in ticks.index:
                    return Tick(ticks.loc[time])
                await self.preload_ticks(symbol=symbol)
                tick = self.preloaded_ticks[symbol].loc[time]
                return Tick(tick)

            elif self.use_terminal and self.preload is False:
                time = datetime.fromtimestamp(time, tz=UTC)
                tick = await self.mt5.copy_ticks_from(symbol, time, 1, CopyTicks.ALL)
                return Tick(tick[-1]) if tick is not None else None
            else:
                tick = self.prices[symbol].loc[time]
                return Tick(tick)
        except Exception as exe:
            logger.error("Error Getting Price Tick: %s", exe)

    @error_handler
    async def check_order(self, *, ticket: int):
        """ "
        Check if the order has reached its take profit or stop loss levels and close the order if it has.
        Checks only **OrderType.BUY** and **OrderType.SELL** orders that have reached their take profit or stop loss levels.

        Args:
            ticket (int): Order ticket
        """
        order = self.orders[ticket]
        order_type, symbol = order.type, order.symbol
        tick = await self.get_price_tick(symbol=symbol, time=self.cursor.time)
        tp, sl = order.tp, order.sl

        if not (tp and sl):
            return

        deal = {
            "ticket": random.randint(100_000_000, 199_999_999),
            "order": ticket,
            "symbol": symbol,
            "commission": 0,
            "swap": 0,
            "position_id": ticket,
            "fee": 0,
            "time": self.cursor.time,
            "time_msc": self.cursor.time * 1000,
            "price": tick.bid,
            "type": DealType(order_type),
            "reason": DealReason.EXPERT,
            "entry": DealEntry.OUT,
            "profit": 0,
        }

        match order_type:
            case OrderType.BUY:
                if tick.bid >= tp or tick.bid <= sl:
                    res = await self.close_position(ticket=ticket)
                    if res:
                        pos = self.positions.get(ticket)
                        deal.update({"profit": pos.profit, "volume": pos.volume})
                        self.deals[deal["ticket"]] = TradeDeal((deal.get(k, 0) for k in TradeDeal.__match_args__))

            case OrderType.SELL:
                if tick.ask <= tp or tick.ask >= sl:
                    res = await self.close_position(ticket=ticket)
                    if res:
                        pos = self.positions.get(ticket)
                        deal.update({"profit": pos.profit, "volume": pos.volume})
                        self.deals[deal["ticket"]] = TradeDeal((deal.get(k, 0) for k in TradeDeal.__match_args__))
            case _:
                ...

    def check_account(self):
        """Checks an account status. This method is called at each iteration to check if the account has burned out."""
        account = self._account
        level = account.margin_level if account.margin_so_mode == AccountStopOutMode.PERCENT else account.margin_so_call
        if level < account.margin_so_call and level != 0 and account.equity < 0:
            logger.critical("Account has burned out!!! Please top up to continue trading")
            self.stop_testing = True

    async def check_position(self, *, ticket: int):
        """
        Update the profit of an open position based on the current price of the symbol. It is called by the
        tracker to update the profit of open positions.

        Args:
            ticket (int): Position ticket
        """
        pos = self.positions[ticket]
        order_type, symbol, volume, price_open, prev_profit = (
            pos.type,
            pos.symbol,
            pos.volume,
            pos.price_open,
            pos.profit,
        )
        tick = await self.get_price_tick(symbol=symbol, time=self.cursor.time)
        price_current = tick.bid if order_type == OrderType.BUY else tick.ask
        profit = await self.order_calc_profit(
            action=order_type, symbol=symbol, volume=volume, price_open=price_open, price_close=price_current
        )
        kwargs = dict(price_current=price_current, time_update=self.cursor.time)
        kwargs.update(profit=profit) if profit is not None else ...
        self.positions.update(ticket=pos.ticket, **kwargs)
        await self.check_order(ticket=ticket)

    @error_handler_sync
    async def close_position_manually(self, *, ticket: int):
        """Close a position manually without. Usually at the end of testing."""
        res = await self.close_position(ticket=ticket)
        if not res:
            return
        position = self.positions.get(ticket)
        order_ticket = random.randint(800_000_000, 899_999_999)
        deal_ticket = random.randint(100_000_000, 199_999_999)
        time = position.time_update
        time_msc = position.time_update_msc
        order_type = OrderType.BUY if position.type == OrderType.SELL else OrderType.SELL
        order = {
            "position_id": position.ticket,
            "ticket": order_ticket,
            "comment": "",
            "external_id": "",
            "time_setup": time,
            "time_setup_msc": time_msc,
            "time_done": time,
            "time_done_msc": time_msc,
            "type": order_type,
            "symbol": position.symbol,
            "sl": position.sl,
            "tp": position.tp,
            "price_current": position.price_current,
            "reason": OrderReason.EXPERT,
            "volume_initial": position.volume,
        }

        # TODO: calculate commission and swap if possible or necessary
        deal = {
            "ticket": deal_ticket,
            "position_id": position.ticket,
            "order": order_ticket,
            "symbol": position.symbol,
            "time": time,
            "profit": position.profit,
            "time_msc": time_msc,
            "volume": position.volume,
            "price": position.price_current,
            "type": DealType(order_type),
            "reason": DealReason.EXPERT,
            "entry": DealEntry.OUT,
            "comment": "",
            "external_id": "",
        }
        order = TradeOrder((order.get(k, 0) for k in TradeOrder.__match_args__))
        deal = TradeDeal((deal.get(k, 0) for k in TradeDeal.__match_args__))
        self.orders[order.ticket] = order
        self.deals[deal.ticket] = deal

    async def close_position(self, *, ticket: int) -> bool:
        """
        Close an open position for the trading account using the position ticket.

        Args:
            ticket: Position ticket

        Returns:
            bool: True if the position is closed successfully, False otherwise
        """
        try:
            position = self.positions[ticket]
            margin = self.positions.get_margin(ticket=ticket)
            self.positions.delete_margin(ticket=ticket)
            self.positions.close(ticket=ticket)
            self.orders.update(ticket=ticket, time_done=self.cursor.time, time_done_msc=self.cursor.time * 1000)
            profit = sum([position.profit for position in self.positions.open_positions])
            profit = round(profit, self._account.currency_digits)
            self.update_account(gain=position.profit, profit=profit, margin=-margin)
            return True
        except Exception as exe:
            logger.error("Error Closing Position %d: %s", ticket, exe)
            return False

    @error_handler(response=False)
    def modify_stops(self, *, ticket: int, sl: int, tp: int) -> bool:
        """
        Modify the stop loss and take profit levels of an open position.

        Args:
            ticket (int): Position ticket
            sl (int): stop loss level
            tp (int): Take profit level

        Returns:
            bool: True if the stops are modified successfully, False otherwise
        """
        self.positions.update(
            ticket=ticket, sl=sl, tp=tp, time_update=self.cursor.time, time_update_msc=self.cursor.time * 1000
        )
        return True

    def update_account(self, *, profit: float = None, margin: float = 0, gain: float = 0):
        """
        Update the account. This method is protected by thread lock.

        Args:
            profit (float): The current profit of one or more open positions. Can be positive or negative.
            margin (float): The margin set aside for a trade. It is released when the trade is closed.
            gain (gain): The gain realized when the trade is closed.
        """
        self.account_lock.acquire()
        try:
            self._account.balance += round(gain, self._account.currency_digits)
            self._account.profit = (
                round(profit, self._account.currency_digits) if profit is not None else self._account.profit
            )
            self._account.equity = self._account.balance + self._account.profit
            self._account.margin += round(margin, self._account.currency_digits)
            self._account.margin_free = round(
                self._account.equity - self._account.margin, self._account.currency_digits
            )
            self._account.balance = round(self._account.balance, self._account.currency_digits)
            self._account.equity = round(self._account.equity, self._account.currency_digits)
            self._account.margin = round(self._account.margin, self._account.currency_digits)
            self._account.margin_free = round(self._account.margin_free, self._account.currency_digits)
            self._account.profit = round(self._account.profit, self._account.currency_digits)

            if self._account.margin == 0:
                self._account.margin_level = 0
            else:
                mode = self._account.margin_so_mode
                level = self._account.equity / self._account.margin * 100
                self._account.margin_level = level if mode == AccountStopOutMode.PERCENT else self._account.margin_free
        except Exception as exe:
            logger.critical("Error Updating Account: %s", exe)

        finally:
            self.account_lock.release()

    def deposit(self, *, amount: float):
        """Make deposit to the trading account"""
        self.update_account(gain=amount)

    def withdraw(self, *, amount: float):
        """Make a withdrawal from the trading account. You can not withdraw more than what you have"""
        assert amount <= self._account.balance, "Insufficient funds"
        self.update_account(gain=-amount)

    @error_handler
    async def setup_account(self, **kwargs):
        """Setup the trading account before the begining of a backtesting session.

        Args:
            (**kwargs, Any): Attributes for the backetest account object can be set here.
        """
        kwargs = {**self.account_info, **kwargs}
        default = {
            "profit": self._account.profit,
            "margin": self._account.margin,
            "equity": self._account.equity,
            "margin_free": self._account.margin_free,
            "margin_level": self._account.margin_level,
            "balance": self._account.balance,
            **{k: v for k, v in kwargs.items() if k in self._account.__match_args__},
        }

        if self.use_terminal:
            acc_info = await self.mt5.account_info()
            default = {**acc_info._asdict(), **default}

        self._account.set_attrs(**default)
        self.update_account()

    @error_handler_sync
    def setup_account_sync(self, **kwargs):
        """Set up the backtesting account in sync mode"""
        kwargs = {**self.account_info, **kwargs}
        default = {
            "profit": self._account.profit,
            "margin": self._account.margin,
            "equity": self._account.equity,
            "margin_free": self._account.margin_free,
            "margin_level": self._account.margin_level,
            "balance": self._account.balance,
            **{k: v for k, v in kwargs.items() if k in self._account.__match_args__},
        }

        if self.use_terminal:
            acc_info = self.mt5._account_info()
            default = {**acc_info._asdict(), **default}

        self._account.set_attrs(**default)
        self.update_account()

    @cached_property
    def prices(self) -> dict[str, DataFrame]:
        """Get the prices for instruments used in the backtesting. This class is called when the use_terminal option
        is set to False and trading data is provided in the data attribute. It makes sure that there is a price for each
        symbol for every second covered in the backtesting range, by reindexing the price ticks using the backtesting
        time span and filling up missing data using the nearest method.
        This method returns a dictionaries of dataframe containing the prices for each symbol.
        It's cached and there computed only once per backtesting session.

        Returns:
             dict[str, DataFrame]: A dictionary mapping dataframe of prices to symbols.
        """
        prices = {}
        for symbol in self._data.ticks.keys():
            res = self._data.ticks[symbol]
            res = pd.DataFrame(res)
            res.drop_duplicates(subset=["time"], keep="last", inplace=True)
            res.set_index("time", inplace=True, drop=False)
            res = res.reindex(self.span, copy=True, method="nearest")  # fill in missing values with NaN
            prices[symbol] = res
        return prices

    @cached_property
    def ticks(self) -> dict[str, DataFrame]:
        """Similar to prices above, but returns prices exactly as they are without reindexing and filling up.

        Returns:
            dict[str, DataFrame]: A dictionary mapping symbols to dataframes of ticks.
        """
        ticks = {}
        for symbol in self._data.ticks.keys():
            res = self._data.ticks[symbol]
            res = pd.DataFrame(res)
            ticks[symbol] = res
        return ticks

    @cached_property
    def rates(self) -> dict[str, dict[int, DataFrame]]:
        """This property is useful when backtesting with the use_terminal option set to false. It returns a nested dict
        that maps symbols to a dict mapping timeframes to rates. The timeframes are mapped using their integer values.

        Returns:
            dict[str, dict[int, DataFrame]]: A dictionary containing the symbol rates.
        """
        rates = {}
        for symbol in self._data.rates.keys():
            for timeframe in self._data.rates[symbol].keys():
                res = self._data.rates[symbol][timeframe]
                res = pd.DataFrame(res)
                rates.setdefault(symbol, {})[timeframe] = res
        return rates

    @cached_property
    def symbols(self) -> dict[str, SymbolInfo]:
        """A dictionary of symbols and SymbolInfo object. Used when use_terminal is set to false.

        Returns:
            dict[str, SymbolInfo]
        """
        symbols = {}
        for symbol, info in self._data.symbols.items():
            symbols[symbol] = SymbolInfo((info.get(key) for key in SymbolInfo.__match_args__))
        return symbols

    @error_handler
    async def order_send(self, *, request: dict, use_terminal=False) -> OrderSendResult:
        """Simulates the sending of an order to the broker. An OrderSendResult is object is created at the end of this
        operation as would be created if it was done in live trading. When an order is successful a positions object is
        created, an order and deal object is created as well. When use_terminal is set to true the margin and profit
        are calculated by sending to the broker. This increases accuracy but slows down the backtester. Check order is
        called to make sure the order is valid and would go through if it was a live trade.

        Args:
            request (dict): The order request as a dict.
            use_terminal (bool): A flag to override the use_terminal attribute. If true, the terminal will
            be used even if the use_terminal attribute is True.

        Returns:
            OrderSendResult: An object containing the result of the order send operation.
        """
        use_terminal = self.use_terminal or use_terminal
        osr = {
            "retcode": 10013,
            "comment": "Invalid request",
            "request": TradeRequest(request.get(k, (0 if k != "comment" else "")) for k in TradeRequest.__match_args__),
        }
        current_tick = await self.get_price_tick(symbol=request.get("symbol"), time=self.cursor.time)
        if current_tick is None:
            osr["comment"] = "Market is closed"
            osr["retcode"] = 10018
            return OrderSendResult((osr.get(k, 0) for k in OrderSendResult.__match_args__))

        trade_order = {
            "external_id": "",
            "comment": "",
            **{k: v for k, v in request.items() if k in TradeOrder.__match_args__},
        }
        order_type, symbol = request.get("type"), request.get("symbol", "")
        action, position_id = request.get("action"), request.get("position")
        sl, tp, volume, symbol = (request.get("sl"), request.get("tp"), request.get("volume"), request.get("symbol"))
        order_type = OrderType(order_type)
        current_position = self.positions.get(position_id)
        order_ticket = random.randint(800_000_000, 899_999_999)
        deal_ticket = random.randint(100_000_000, 199_999_999)

        # closing an order by an opposite order using a position ticket and Deal action
        if action == TradeAction.DEAL and current_position and order_type.opposite == current_position.type:
            res = await self.close_position(ticket=current_position.ticket)
            if res:
                price_current = current_tick.ask if order_type == OrderType.BUY else current_tick.bid
                trade_order.update(
                    {
                        "position_id": current_position.ticket,
                        "ticket": order_ticket,
                        "time_setup": current_tick.time,
                        "time_setup_msc": current_tick.time_msc,
                        "time_done": current_tick.time,
                        "time_done_msc": current_tick.time_msc,
                        "type": order_type,
                        "symbol": symbol,
                        "sl": current_position.sl,
                        "tp": current_position.tp,
                        "price_current": price_current,
                        "reason": OrderReason.EXPERT,
                        "volume_initial": current_position.volume,
                    }
                )

                # TODO: calculate commission and swap if possible or necessary
                deal = {
                    "ticket": deal_ticket,
                    "position_id": current_position.ticket,
                    "order": order_ticket,
                    "symbol": symbol,
                    "time": current_tick.time,
                    "profit": current_position.profit,
                    "time_msc": current_tick.time_msc,
                    "volume": current_position.volume,
                    "price": price_current,
                    "type": DealType(order_type),
                    "reason": DealReason.EXPERT,
                    "entry": DealEntry.OUT,
                    "comment": "",
                    "external_id": "",
                }

                order = TradeOrder((trade_order.get(k, 0) for k in TradeOrder.__match_args__))
                self.orders[order.ticket] = order
                deal = TradeDeal((deal.get(k, 0) for k in TradeDeal.__match_args__))
                self.deals[deal.ticket] = deal
                osr.update(
                    {"comment": "Request completed", "retcode": 10009, "order": order_ticket, "deal": deal_ticket}
                )
            return OrderSendResult((osr.get(k, 0) for k in OrderSendResult.__match_args__))

        if action == TradeAction.SLTP and current_position:
            check = await self.order_check(request=request, use_terminal=use_terminal)
            if check.retcode != 0:
                osr = {"retcode": check.retcode, "comment": check.comment, "request": check.request}
                return OrderSendResult((osr.get(k, 0) for k in OrderSendResult.__match_args__))
            res = self.modify_stops(ticket=position_id, sl=sl, tp=tp)
            if res:
                osr.update(
                    {"comment": "Request completed", "retcode": 10009, "order": order_ticket, "deal": deal_ticket}
                )
            return OrderSendResult((osr.get(k, 0) for k in OrderSendResult.__match_args__))

        if action == TradeAction.DEAL and order_type in (OrderType.BUY, OrderType.SELL):
            check = await self.order_check(request=request, use_terminal=use_terminal)
            if check.retcode != 0:
                osr = {"retcode": check.retcode, "comment": check.comment, "request": check.request}
                return OrderSendResult((osr.get(k, 0) for k in OrderSendResult.__match_args__))

            price = current_tick.ask if order_type == OrderType.BUY else current_tick.bid
            position = {
                "ticket": order_ticket,
                "symbol": symbol,
                "volume": volume,
                "price_open": price,
                "price_current": price,
                "type": order_type,
                "profit": 0,
                "reason": PositionReason.EXPERT,
                "identifier": order_ticket,
                "sl": sl,
                "tp": tp,
                "time": current_tick.time,
                "time_msc": current_tick.time_msc,
                "time_update": current_tick.time,
                "time_update_msc": current_tick.time_msc,
            }

            deal = {
                "ticket": deal_ticket,
                "order": order_ticket,
                "symbol": symbol,
                "commission": 0,
                "swap": 0,
                "position_id": order_ticket,
                "fee": 0,
                "time": current_tick.time,
                "time_msc": current_tick.time_msc,
                "volume": volume,
                "price": price,
                "type": DealType(order_type),
                "reason": DealReason.EXPERT,
                "entry": DealEntry.IN,
                "profit": 0,
            }

            # ToDo: set time_expiration based on order_type_time
            trade_order.update(
                {
                    "ticket": order_ticket,
                    "symbol": symbol,
                    "volume": volume,
                    "price": price,
                    "price_current": price,
                    "sl": sl,
                    "time_setup_msc": current_tick.time_msc,
                    "tp": tp,
                    "price_open": price,
                    "type": order_type,
                    "time_setup": current_tick.time,
                    "volume_current": volume,
                    "volume_initial": volume,
                    "position_id": order_ticket,
                }
            )

            pos = TradePosition((position.get(k, 0) for k in TradePosition.__match_args__))
            order = TradeOrder((trade_order.get(k, 0) for k in TradeOrder.__match_args__))
            deal = TradeDeal((deal.get(k, 0) for k in TradeDeal.__match_args__))
            self.deals[deal_ticket] = deal
            self.positions[order.ticket] = pos
            self.orders[order.ticket] = order
            osr.update(
                {
                    "comment": "Request completed",
                    "retcode": 10009,
                    "order": order_ticket,
                    "price": price,
                    "volume": volume,
                    "bid": current_tick.bid,
                    "ask": current_tick.ask,
                    "deal": deal_ticket,
                }
            )
            margin = await self.order_calc_margin(
                action=action, symbol=symbol, volume=volume, price=price, use_terminal=use_terminal
            )
            self.positions.set_margin(ticket=order_ticket, margin=margin)
            self.update_account(margin=margin)
            return OrderSendResult((osr.get(k, 0) for k in OrderSendResult.__match_args__))

    @error_handler
    async def order_check(self, *, request: dict, use_terminal: bool = False) -> OrderCheckResult:
        """Checks the order before placing it. If use_terminal, the order is checked with the broker, but the entire result
        is not used. Details such as balance, profit, equity, margin, and margin level are calculated by the backtester.

        Args:
            request (dict): The order request as a dict.
            use_terminal (bool): A flag to override the use_terminal attribute. If true, the terminal will used.

        Returns:
            OrderCheckResult: The result of the order check.
        """
        use_terminal = self.use_terminal or use_terminal
        ocr = {
            "retcode": 10013,
            "balance": 0,
            "profit": 0,
            "margin": 0,
            "equity": 0,
            "margin_free": 0,
            "margin_level": 0,
            "comment": "Invalid request",
            "request": TradeRequest(request.get(k, (0 if k != "comment" else "")) for k in TradeRequest.__match_args__),
        }

        action, symbol, volume = (request.get("action"), request.get("symbol"), request.get("volume"))
        price, order_type, position_id = (request.get("price"), request.get("type"), request.get("position"))

        # check margin and confirm order can go through for a deal action and buy or sell order type
        if action == TradeAction.DEAL and order_type in (OrderType.BUY, OrderType.SELL) and position_id is None:
            margin = await self.order_calc_margin(
                action=action, symbol=symbol, volume=volume, price=price, use_terminal=use_terminal
            )
            if margin is None:
                return OrderCheckResult((ocr.get(k, 0) for k in OrderCheckResult.__match_args__))

            used_margin = self._account.margin + margin
            free_margin = self._account.margin_free - margin

            level = self._account.equity / used_margin * 100 if used_margin else float("inf")
            margin_level = level if self._account.margin_so_mode == AccountStopOutMode.PERCENT else free_margin
            ocr.update({"margin_level": margin_level, "margin": margin, "margin_free": free_margin})

            # check if the account has enough money
            if margin_level < self._account.margin_so_call:
                ocr["retcode"] = 10019
                ocr["comment"] = "No money"
                return OrderCheckResult((ocr.get(k, 0) for k in OrderCheckResult.__match_args__))

        # check if the stops level is valid
        sym = await self.get_symbol_info(symbol=symbol)
        sl, tp = request.get("sl"), request.get("tp")
        current_price = price
        if tp and sl:
            if action == TradeAction.SLTP:
                pos = self.positions.get(request.get("position"))
                sym = await self.get_symbol_info(symbol=pos.symbol)
                current_tick = sym or await self.get_price_tick(pos.symbol, self.cursor.time)
                current_price = current_tick.bid if pos.type == OrderType.BUY else current_tick.ask

            min_sl = min(sl, tp)
            dsl = abs(current_price - min_sl) / sym.point
            tsl = sym.trade_stops_level + sym.spread
            if dsl < tsl:
                ocr["retcode"] = 10016
                ocr["comment"] = "Invalid stops"
                return OrderCheckResult((ocr.get(k, 0) for k in OrderCheckResult.__match_args__))

            elif action == TradeAction.SLTP:
                ocr["comment"] = "Done"
                ocr["retcode"] = 0
                return OrderCheckResult((ocr.get(k, 0) for k in OrderCheckResult.__match_args__))

        if use_terminal or self.use_terminal:
            ocr_t = await self.mt5.order_check(request)
            if ocr_t.retcode in (10013, 10014):
                return ocr_t
        elif action == TradeAction.DEAL and order_type in (OrderType.BUY, OrderType.SELL):
            # check volume
            if volume < sym.volume_min or volume > sym.volume_max:
                ocr["retcode"] = 10014
                ocr["comment"] = "Invalid volume"
                return OrderCheckResult((ocr.get(k, 0) for k in OrderCheckResult.__match_args__))

        ocr.update(
            {
                "balance": self._account.balance,
                "profit": self._account.profit,
                "equity": self._account.equity,
                "comment": "Done",
                "retcode": 0,
            }
        )

        return OrderCheckResult((ocr.get(k, 0) for k in OrderCheckResult.__match_args__))

    @error_handler
    async def get_terminal_info(self) -> TerminalInfo:
        """Get the terminal information

        Returns:
            TerminalInfo: The terminal information
        """
        if self.use_terminal:
            res = await self.mt5.terminal_info()
            return res
        return TerminalInfo(self._data.terminal)

    @error_handler
    async def get_version(self) -> tuple[int, int, str]:
        """Get the version of the terminal.

        Returns:
            tuple[int, int, str]: The version of the terminal
        """
        if self.use_terminal:
            res = await self.mt5.version()
            return res
        return self._data.version

    @error_handler
    async def get_symbols_total(self) -> int:
        """Get the total number of symbols available in the terminal.

        Returns:
            int: The total number of symbols available.
        """
        if self.use_terminal:
            syms = await self.mt5.symbols_total()
            return syms
        return len(self.symbols)

    @error_handler
    async def get_symbols(self, *, group: str = "") -> tuple[SymbolInfo, ...]:
        """Get the symbols available in the terminal. Filter by group if provided.

        Args:
            group (str): The group to filter by (default is "")

        Returns:
            tuple[SymbolInfo, ...]: A tuple of symbol information
        """
        if self.use_terminal:
            syms = await self.mt5.symbols_get(group=group)
            return syms
        return tuple(list(self.symbols.values()))

    @error_handler_sync
    def get_account_info(self) -> AccountInfo:
        """Get the account information

        Returns:
            AccountInfo: The account information
        """
        return AccountInfo(self._account.asdict().values())

    @error_handler
    async def get_symbol_info_tick(self, *, symbol: str) -> Tick:
        """Get the price tick for a symbol at the current time

        Args:
            symbol (str): The symbol

        Returns:
            Tick: The price tick
        """
        tick = await self.get_price_tick(symbol=symbol, time=self.cursor.time)
        return tick

    @async_cache
    async def _symbol_info(self, *, symbol: str) -> SymbolInfo:
        if self.use_terminal:
            info = await self.mt5.symbol_info(symbol)
        else:
            info = self.symbols[symbol]
        return info

    @error_handler
    async def get_symbol_info(self, *, symbol: str) -> SymbolInfo:
        """Get the symbol information

        Args:
            symbol (str): The symbol to get information for

        Returns:
            SymbolInfo: The symbol information
        """
        if self.use_terminal:
            info = await self._symbol_info(symbol=symbol)
        else:
            info = self.symbols[symbol]
        tick = await self.get_symbol_info_tick(symbol=symbol)

        info = info._asdict() | {
            "bid": tick.bid,
            "bidhigh": tick.bid,
            "bidlow": tick.bid,
            "ask": tick.ask,
            "askhigh": tick.ask,
            "asklow": tick.bid,
            "last": tick.last,
            "volume_real": tick.volume_real,
        }
        return SymbolInfo((info.get(key) for key in SymbolInfo.__match_args__))

    @error_handler
    async def get_rates_from(
        self, *, symbol: str, timeframe: TimeFrame, date_from: datetime | float, count: int
    ) -> np.ndarray:
        """Get rates from a specific date to the current date. Used by the backtester to get rates for a symbol

        Args:
            symbol (str): The symbol to get rates for
            timeframe (TimeFrame): The timeframe of the rates
            date_from (datetime | float): The date from which to get the rates
            count (int): The number of rates to get

        Returns:
            np.ndarray: An array of rates
        """
        date_from = (
            date_from.astimezone(tz=UTC)
            if isinstance(date_from, datetime)
            else datetime.fromtimestamp(date_from, tz=UTC)
        )
        if self.use_terminal:
            rates = await self.mt5.copy_rates_from(symbol, timeframe, date_from, count)
            return rates

        rates = self.rates[symbol][timeframe]
        start = int(date_from.timestamp())
        start = round_down(start, timeframe.seconds)
        rates = rates[rates.time <= start].iloc[-count:]
        return np.fromiter((tuple(i) for i in rates.iloc), dtype=self.get_dtype(df=rates))

    @error_handler
    async def get_rates_from_pos(self, *, symbol: str, timeframe: TimeFrame, start_pos: int, count: int) -> np.ndarray:
        """Get a number of rates counting from a specific position. With position zero being the current time.

        Args:
            symbol (str): The symbol to get rates for
            timeframe (TimeFrame): The timeframe of the rates
            start_pos (int): The position to start from
            count (int): The number of rates to get

        Returns:
            np.ndarray: An array of rates
        """
        if self.use_terminal:
            current_time = self.cursor.time if start_pos == 0 else self.cursor.time - start_pos * timeframe.seconds
            current_time = round_up(current_time, timeframe.seconds)
            start = datetime.fromtimestamp(current_time, tz=UTC)
            rates = await self.mt5.copy_rates_from(symbol, timeframe, start, count)
            return rates

        rates = self.rates[symbol][timeframe]

        # the current time rounded up to a multiple of the timeframe in seconds and then subtracted by the start_pos
        # multiplied by the timeframe in seconds gives the time of the last candlestick in the range when using
        # copy_rates_from_pos
        end = int(round_down(self.cursor.time, timeframe.seconds)) - start_pos * timeframe.seconds
        start = end - count * timeframe.seconds
        rates = rates[(rates.time > start) & (rates.time <= end)]
        return np.fromiter((tuple(i) for i in rates.iloc), dtype=self.get_dtype(df=rates))

    @error_handler
    async def get_rates_range(
        self, *, symbol: str, timeframe: TimeFrame, date_from: datetime | float, date_to: datetime | float
    ) -> np.ndarray:
        """Get rates within a specific date range. Used by the backtester to get rates for a symbol

        Args:
            symbol (str): The symbol to get rates for
            timeframe (TimeFrame): The timeframe of the rates
            date_from (datetime | float): The date from which to get the rates
            date_to (datetime | float): The date to which to get the rates

        Returns:
            np.ndarray: An array of rates
        """
        date_from = (
            date_from.astimezone(tz=UTC)
            if isinstance(date_from, datetime)
            else datetime.fromtimestamp(date_from, tz=UTC)
        )
        date_to = (
            date_to.astimezone(tz=UTC) if isinstance(date_to, datetime) else datetime.fromtimestamp(date_to, tz=UTC)
        )
        if self.use_terminal:
            rates = await self.mt5.copy_rates_range(symbol, timeframe, date_from, date_to)
            return rates

        rates = self.rates[symbol][timeframe]
        start = round_up(int(date_from.timestamp()), timeframe.seconds)
        end = round_up(int(date_to.timestamp()), timeframe.seconds)
        rates = rates[(rates.time >= start) & (rates.time <= end)]
        return np.fromiter((tuple(i) for i in rates.iloc), dtype=self.get_dtype(df=rates))

    @error_handler
    async def get_ticks_from(
        self, *, symbol: str, date_from: datetime | float, count: int, flags: CopyTicks = CopyTicks.ALL
    ) -> np.ndarray:
        """Get a specified number of ticks counting from a specific date.
        Args:
            symbol (str): The symbol to get ticks for
            date_from (datetime | float): The date from which to get the ticks
            count (int): The number of ticks to get
            flags (CopyTicks): The flags to use when getting the ticks

        Returns:
            np.ndarray: An array of ticks
        """
        date_from = (
            date_from.astimezone(tz=UTC)
            if isinstance(date_from, datetime)
            else datetime.fromtimestamp(date_from, tz=UTC)
        )
        if self.use_terminal:
            ticks = await self.mt5.copy_ticks_from(symbol, date_from, count, flags)
            return ticks

        ticks = self.ticks[symbol]
        start = int(date_from.timestamp())
        rates = ticks[ticks.time <= start].iloc[-count:]
        return np.fromiter((tuple(i) for i in rates.iloc), dtype=self.get_dtype(df=rates))

    @error_handler
    async def get_ticks_range(
        self, *, symbol: str, date_from: datetime | float, date_to: datetime | float, flags: CopyTicks = CopyTicks.ALL
    ) -> np.ndarray:
        """Get ticks within a specific date range.

        Args:
            symbol (str): The symbol to get ticks for
            date_from (datetime | float): The date from which to get the ticks
            date_to (datetime | float): The date to which to get the ticks
            flags (CopyTicks): The flags to use when getting the ticks

        Returns:
            np.ndarray: An array of ticks
        """
        date_from = (
            date_from.astimezone(tz=UTC)
            if isinstance(date_from, datetime)
            else datetime.fromtimestamp(date_from, tz=UTC)
        )
        date_to = (
            date_to.astimezone(tz=UTC) if isinstance(date_to, datetime) else datetime.fromtimestamp(date_to, tz=UTC)
        )
        if self.use_terminal:
            ticks = await self.mt5.copy_ticks_range(symbol, date_from, date_to, flags)
            return ticks

        ticks = self.ticks[symbol]
        start = int(date_from.timestamp())
        end = int(date_to.timestamp())
        rates = ticks[(ticks.time >= start) & (ticks.time <= end)]
        return np.fromiter((tuple(i) for i in rates.iloc), dtype=self.get_dtype(df=ticks))

    @error_handler
    async def order_calc_margin(
        self,
        *,
        action: Literal[OrderType.BUY, OrderType.SELL],
        symbol: str,
        volume: float,
        price: float,
        use_terminal: bool = None,
    ):
        """Calculate the margin required for a trade.

        Args:
            action (Literal[OrderType.BUY, OrderType.SELL]): Type of order
            symbol (str): Symbol name
            volume (float): Volume of the trade
            price (float): The price at which the trade is opened
            use_terminal (bool): A flag to override the use_terminal attribute. If true, the terminal will be used
             even if the use_terminal attribute is True.

        Returns:
            float: The margin required for the trade
        """
        use_terminal = use_terminal if use_terminal is not None else self.use_terminal
        if use_terminal:
            return await self.mt5.order_calc_margin(action, symbol, volume, price)

        sym = self.symbols.get(symbol)
        if sym is None and self.use_terminal:
            sym = await self._symbol_info(symbol=symbol)
        margin = (volume * sym.trade_contract_size * price) / (self._account.leverage / (sym.margin_initial or 1))
        return round(margin, self._account.currency_digits)

    @error_handler
    async def order_calc_profit(
        self,
        *,
        action: Literal[OrderType.BUY, OrderType.SELL],
        symbol: str,
        volume: float,
        price_open: float,
        price_close: float,
        use_terminal=None,
    ):
        """
        Calculate the profit for a trade.

        Args:
            action (Literal[OrderType.BUY, OrderType.SELL]): Type of order
            symbol (str): Symbol name
            volume (float): Volume of the trade
            price_open (float): The price at which the trade is opened
            price_close (float): The price at which the trade is closed
            use_terminal (bool): A flag to override the use_terminal attribute. If true, the terminal will be used
             even if the use_terminal attribute is True.

        Returns:
            float: The profit of the trade
        """
        use_terminal = use_terminal if use_terminal is not None else self.use_terminal

        if use_terminal:
            return await self.mt5.order_calc_profit(action, symbol, volume, price_open, price_close)

        sym = self.symbols.get(symbol)
        if sym is None and self.use_terminal:
            sym = await self._symbol_info(symbol=symbol)
        profit = (
            volume
            * sym.trade_contract_size
            * ((price_close - price_open) if action == OrderType.BUY else (price_open - price_close))
        )
        return round(profit, self._account.currency_digits)

    @error_handler_sync
    def get_orders_total(self) -> int:
        """Get the total number of pending orders.

        Returns:
            int: Total number of pending orders
        """
        return 0

    @error_handler_sync
    def get_orders(self, *, symbol: str = "", group: str = "", ticket: int = None) -> tuple[TradeOrder, ...]:
        """Get pending orders from the terminal history. This has to do with pending orders, which this backtester
        doesn't support yet.

        Args:
            symbol: Symbol name
            group: Group name
            ticket: Order ticket

        Returns:
            tuple[TradeOrder, ...]: Pending orders
        """
        if symbol and group and ticket:
            return tuple()
        return ()

    @error_handler_sync
    def get_positions_total(self) -> int:
        """Get the total number of open positions.

        Returns:
            int: Total number of open positions
        """
        return self.positions.positions_total()

    @error_handler_sync
    def get_positions(self, *, symbol: str = None, group: str = None, ticket: int = None) -> tuple[TradePosition, ...]:
        """Get open positions from the terminal history.

        Args:
            symbol: The symbol name
            group: Group argument to filter by
            ticket: Position ticket

        Returns:
            tuple[TradePosition, ...]: Open positions
        """
        return self.positions.positions_get(ticket=ticket, symbol=symbol, group=group)

    @error_handler_sync
    def get_history_orders_total(self, *, date_from: datetime | float, date_to: datetime | float) -> int:
        """Get the total number of orders in the terminal history.

        Args:
            date_from: The start date of the history

            date_to: The end date of the history

        Returns:
            int: Total number of orders in the history
        """
        return self.orders.history_orders_total(date_from=date_from, date_to=date_to)

    @error_handler_sync
    def get_history_orders(
        self,
        *,
        date_from: datetime | float = None,
        date_to: datetime | float = None,
        group: str = "",
        ticket: int = None,
        position: int = None,
    ) -> tuple[TradeOrder, ...]:
        """Get orders from the terminal history.

        Args:
            date_from: Date from which to start the history
            date_to: Date to which to end the history
            group: group keyword to filter by
            ticket: ticket id to filter by
            position: position id to filter by

        Returns:
            tuple[TradeOrder, ...]: Orders in the history
        """
        return self.orders.history_orders_get(
            date_from=date_from, date_to=date_to, group=group, ticket=ticket, position=position
        )

    @error_handler_sync
    def get_history_deals_total(self, *, date_from: datetime | float, date_to: datetime | float) -> int:
        """Get the total number of deals in the terminal history.

        Args:
            date_from: Date from which to start the history
            date_to: Date to which to end the history

        Returns:
            int: Total number of deals in the history
        """
        return self.deals.history_deals_total(date_from=date_from, date_to=date_to)

    @error_handler_sync
    def get_history_deals(
        self,
        *,
        date_from: datetime | float = None,
        date_to: datetime | float = None,
        group: str = None,
        position: int = None,
        ticket: int = None,
    ) -> tuple[TradeDeal, ...]:
        """Get deals from the terminal history.

        Args:
            date_from: Date from which to start the history
            date_to: Date to which to end the history
            group: group keyword to filter by
            position: position id to filter by
            ticket: ticket id to filter by

        Returns:
            tuple[TradeDeal, ...]: Deals in the history
        """
        return self.deals.history_deals_get(
            date_from=date_from, date_to=date_to, group=group, position=position, ticket=ticket
        )


BackTestEngine.__doc__ = """The BackTestEngine class is used to simulate trading strategies on historical data.
        It can accept already saved data or create new data for backtesting on the fly. Ideally only one instance of
        this class should be created per session. By default it is automatically assigned to the global config instance
        during instantiation, replacing any existing backtest engine instance. But this is a configurable behaviour.
        The start and end time can still be specified even when test data is provided. In that case it will be used
        to set the range of the backtest.

        Args:
            data (BackTestData, optional): The data to use for backtesting. Defaults to None.

            speed (int, optional): The speed of the backtest. Defaults to 60 seconds.

            start (float | datetime, optional): The start time of the backtest. Defaults to 0. If a float is passed,
             it is assumed to be a timestamp.

            end (float | datetime, optional): The end time of the backtest. Defaults to 0. If a float is passed,
             it is assumed to be a timestamp.

            restart (bool, optional): Whether to restart the backtest from the beginning. Defaults to True.
             This is useful when resuming a backtest using a saved BackTestData instance.

            use_terminal (bool, optional): Whether to use the terminal for backtesting. Defaults to None. If None,
             it uses the global config setting. If use terminal is true, the backtest engine will use the terminal to
             get price data, compute margins, profit and check order viability. If false, it will use the data
             provided in the BackTestData instance and default algorithm for the calculations

            name (str, optional): The name of the backtest. Defaults to "". If not provided,
             it is generated from the start and end times.

            stop_time (float | datetime, optional): The time to stop the backtest. Defaults to None.
             If a float is passed, it is assumed to be a timestamp. If not given it is assumed to be the end of the backtest range.

            close_open_positions_on_exit (bool, optional): Whether to close all open positions when the backtest
             is stopped. Defaults to True.

            preload (bool, optional): Whether to preload the ticks for the backtest. Defaults to True.

            assign_to_config (bool, optional): Whether to assign the backtest engine to the global config instance.
             Defaults to True.

            account_info (dict, optional): A dictionary of account information to use for the backtest. Defaults to None. Use this to set
             the account information for the backtest.

        Attributes:
            _data (BackTestData): The data used for backtesting. This is the data that is saved to disk when the
             backtest is stopped.

            mt5 (MetaTrader): The MetaTrader instance for the backtest engine.

            config (Config): The global configuration instance.

            name (str): The name of the backtest.

            stop_testing (bool): Whether to stop the backtest.

            use_terminal (bool): Whether to use the terminal for backtesting.

            close_open_positions_on_exit (bool): Whether to close all open positions when the backtest is stopped.

            stop_time (int): The time to stop the backtest.

            preload (bool): Whether to preload the ticks for the backtest.

            preloaded_ticks (dict): A dictionary of preloaded ticks for the backtest.

            account_lock (RLock): A reentrant lock for the account data.

            account_info (dict): A dictionary of account information for the backtest.

        """
